/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.cadmiumjdbc;

import java.sql.RowId;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Custom;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

/**
 * This class provides the primitives related to the <tt>java.sql.RowId</tt> class.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
@PrimitiveProvider
public final class RowIds {

    /**
     * No instance of this class.
     */
    private RowIds() {
    } // end empty constructor

    /**
     * Tests whether the passed rowids are equals.
     * @param ctxt context
     * @param rowid first rowid
     * @param other second rowid
     * @return <tt>true</tt> if the passed rowids are equal,
     *         <tt>false</tt> otherwise
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_rowid_equals(final CodeRunner ctxt,
                                                 final Value rowid,
                                                 final Value other)
        throws Fail.Exception {
        try {
            final RowId ri = (RowId) rowid.asBlock().asCustom();
            final RowId o = (RowId) other.asBlock().asCustom();
            return ri.equals(o) ? Value.TRUE : Value.FALSE;
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_rowid_equals(CodeRunner, Value, Value)'

    /**
     * Returns the value of the passed rowid as a byte array.
     * @param ctxt context
     * @param rowid rowid
     * @return the value of the passed rowid as a byte array
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_rowid_get_bytes(final CodeRunner ctxt,
                                                    final Value rowid)
        throws Fail.Exception {
        try {
            final RowId ri = (RowId) rowid.asBlock().asCustom();
            final Block res = Block.createString(ri.getBytes());
            return Value.createFromBlock(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_rowid_get_bytes(CodeRunner, Value)'

    /**
     * Returns the hash code of the passed rowid.
     * @param ctxt context
     * @param rowid rowid
     * @return the hash code of the passed rowid
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_rowid_hash_code(final CodeRunner ctxt,
                                                    final Value rowid)
        throws Fail.Exception {
        try {
            final RowId ri = (RowId) rowid.asBlock().asCustom();
            final Block res = Block.createCustom(Custom.INT_32_SIZE,
                                                 Custom.INT_32_OPS);
            res.setInt32(ri.hashCode());
            return Value.createFromBlock(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_rowid_hash_code(CodeRunner, Value)'

    /**
     * Converts the passed rowid into a string.
     * @param ctxt context
     * @param rowid rowid
     * @return the passed rowid as a string
     * @throws Fail.Exception if any error occurs
     */
    @Primitive
    public static Value cadmiumjdbc_rowid_to_string(final CodeRunner ctxt,
                                                    final Value rowid)
        throws Fail.Exception {
        try {
            final RowId ri = (RowId) rowid.asBlock().asCustom();
            final Block res = Block.createString(ri.toString());
            return Value.createFromBlock(res);
        } catch (final Exception e) {
            Cadmium.fail(ctxt, e);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'cadmiumjdbc_rowid_to_string(CodeRunner, Value)'

} // end class 'RowIds'
