/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.dbm;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.Enumeration;

import org.w3c.tools.dbm.jdbm;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.util.Misc;

/**
 * This interface defines a Dbm implementation based on JDBM. <br/>
 * This class will be dynamically loaded by a class loader in order to get rid
 * of the dependency to <tt>org.w3c.tools.dbm.jdbm</tt>. More, the load will
 * occur only if explicitly requested by the user (the default is to use a
 * pure-JDK implementation for Dbm).
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
public final class JdbmImplementation implements DbmImplementation {

    /** Suffix for database files. */
    private static final String FILE_SUFFIX = ".db";

    /**
     * Empty constructor.
     */
    public JdbmImplementation() {
    } // end empty constructor

    /**
     * {@inheritDoc}
     */
    public Value open(final CodeRunner ctxt,
                      final Value file,
                      final Value flags,
                      final Value mode)
        throws Fail.Exception, FalseExit {
        try {
            final File f = ctxt.getContext().getRealFile(file.asBlock().asString()
                                                         + JdbmImplementation.FILE_SUFFIX);
            final Value res = createDbm(new jdbm(f.getAbsolutePath()));
            try {
                Misc.setPermissions(f, mode.asLong());
            } catch (final Exception e) {
                // fails silently
            } // end try/catch
            return res;
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Cldbm.error(ctxt, "Can't open/create file");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cldbm.error(ctxt, e.toString());
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'open(CodeRunner, Value, Value, Value)'

    /**
     * {@inheritDoc}
     */
    public Value close(final CodeRunner ctxt,
                       final Value db)
        throws Fail.Exception, FalseExit {
        try {
            get(ctxt, db).db.save();
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            db.asBlock().setCustom(null);
            throw fe;
        } catch (final IOException ioe) {
            // nothing to do ...
        } catch (final Fail.Exception fe) {
            db.asBlock().setCustom(null);
            throw fe;
        } catch (final Exception e) {
            // nothing to do ...
        } // end try/catch
        db.asBlock().setCustom(null);
        return Value.UNIT;
    } // end method 'close(CodeRunner, Value)'

    /**
     * {@inheritDoc}
     */
    public Value fetch(final CodeRunner ctxt,
                       final Value db,
                       final Value key)
        throws Fail.Exception, FalseExit {
        final jdbm database = get(ctxt, db).db;
        try {
            final byte[] res = database.fetch(key.asBlock().getBytes());
            if (res != null) {
                return Value.createFromBlock(Block.createString(res));
            } else {
                Fail.raiseNotFound();
                return Value.UNIT; // never reached
            } // end if/else
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Cldbm.error(ctxt, "dbm_fetch failed");
            return Value.UNIT; // never reached
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cldbm.error(ctxt, e.toString());
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'fetch(CodeRunner, Value, Value)'

    /**
     * {@inheritDoc}
     */
    public Value insert(final CodeRunner ctxt,
                        final Value db,
                        final Value key,
                        final Value content)
        throws Fail.Exception, FalseExit {
        final jdbm database = get(ctxt, db).db;
        try {
            final byte[] res = database.fetch(key.asBlock().getBytes());
            if (res != null) {
                Cldbm.error(ctxt, "Entry already exists");
                return Value.UNIT; // never reached
            } else {
                database.store(key.asBlock().getBytes(),
                               content.asBlock().getBytes(),
                               jdbm.STORE_INSERT);
                return Value.UNIT;
            } // end if/else
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Cldbm.error(ctxt, "dbm_store failed");
            return Value.UNIT; // never reached
        } catch (final Fail.Exception fe) {
            throw fe;
        } catch (final Exception e) {
            Cldbm.error(ctxt, e.toString());
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'insert(CodeRunner, Value, Value, Value)'

    /**
     * {@inheritDoc}
     */
    public Value replace(final CodeRunner ctxt,
                         final Value db,
                         final Value key,
                         final Value content)
        throws Fail.Exception, FalseExit {
        final jdbm database = get(ctxt, db).db;
        try {
            database.store(key.asBlock().getBytes(),
                           content.asBlock().getBytes(),
                           jdbm.STORE_REPLACE);
            return Value.UNIT;
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Cldbm.error(ctxt, "dbm_store failed");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cldbm.error(ctxt, e.toString());
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'replace(CodeRunner, Value, Value, Value)'

    /**
     * {@inheritDoc}
     */
    public Value delete(final CodeRunner ctxt,
                        final Value db,
                        final Value key)
        throws Fail.Exception, FalseExit {
        final jdbm database = get(ctxt, db).db;
        try {
            database.delete(key.asBlock().getBytes());
            return Value.UNIT;
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Cldbm.error(ctxt, "dbm_delete failed");
            return Value.UNIT; // never reached
        } catch (final Exception e) {
            Cldbm.error(ctxt, e.toString());
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'delete(CodeRunner, Value, Value)'

    /**
     * {@inheritDoc}
     */
    public Value firstkey(final CodeRunner ctxt,
                          final Value db)
        throws Fail.Exception {
        final JdbmImplementationDB<jdbm> database = get(ctxt, db);
        final Enumeration e = database.db.keys();
        database.enumeration = e;
        if (!e.hasMoreElements()) {
            Fail.raiseNotFound();
        } // end if
        return Value.createFromBlock(Block.createString((byte[]) e.nextElement()));
    } // end method 'firstkey(CodeRunner, Value)'

    /**
     * {@inheritDoc}
     */
    public Value nextkey(final CodeRunner ctxt,
                         final Value db)
        throws Fail.Exception {
        final Enumeration e = get(ctxt, db).enumeration;
        if ((e == null) || !e.hasMoreElements()) {
            Fail.raiseNotFound();
        } // end if
        return Value.createFromBlock(Block.createString((byte[]) e.nextElement()));
    } // end method 'nextkey(CodeRunner, Value)'

    /**
     * Creates a <i>Dbm.t</i> instance by encapsulating a
     * {@link fr.x9c.cadmium.primitives.dbm.JdbmImplementationDB} object into
     * an abstract value.
     * @param db object to encapsulate - should not be <tt>null</tt>
     * @return passed object encapsulated into an abstract value
     */
    private static Value createDbm(final jdbm db) {
        assert db != null : "null db";
        final Block res = Block.createBlock(1, Block.ABSTRACT_TAG);
        res.setCustom(new JdbmImplementationDB<jdbm>(db));
        return Value.createFromBlock(res);
    } // end method 'createDbm(jdbm)'

    /**
     * Returns the {@link fr.x9c.cadmium.primitives.dbm.JdbmImplementationDB} object
     * encapsulated inside a value.
     * @param ctxt context - should not be <tt>null</tt>
     * @param db value to get <tt>jdbm</tt> object from
     *           - should not be <tt>null</tt>
     * @return the {@link fr.x9c.cadmium.primitives.dbm.JdbmImplementationDB} object
     *         encapsulated inside a value
     * @throws Fail.Exception if <tt>db</tt> has been closed
     */
    @SuppressWarnings("unchecked")
    private static JdbmImplementationDB<jdbm> get(final CodeRunner ctxt,
                                                  final Value db)
        throws Fail.Exception {
        assert ctxt != null : "null ctxt";
        assert db != null : "null db";
        final JdbmImplementationDB<jdbm> res = (JdbmImplementationDB<jdbm>) db.asBlock().asCustom();
        if (res != null) {
            return res;
        } else {
            Cldbm.error(ctxt, "DBM has been closed");
            return null; // never reached
        } // end if/else
    } // end method 'get(CodeRunner, Value)'

} // end class 'JdbmImplementation'
