/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class implements a window listener calling a callback upon either opening,
 * or closing of a window.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public final class CamlWindowAdapter extends WindowAdapter {

    /** Code runner used to execute callback. */
    private final CodeRunner codeRunner;

    /** Slot used to retrieve handler. */
    private final TKSlot slot;

    /** Whether the callback should be called upon opening. */
    private final boolean opening;

    /** Callback identifier. */
    private final int id;

    /**
     * Builds a window listener.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param s used to retrieve handler - should not be <tt>null</tt>
     * @param vis whether the callback should be called upon opening
     * @param id callback identifier 
     */
    public CamlWindowAdapter(final CodeRunner cr, final TKSlot s, final boolean vis, final int i) {
        assert cr != null : "null cr";
        assert s != null : "null s";
        this.codeRunner= cr;
        this.slot = s;
        this.opening = vis;
        this.id = i;
    } // end constructor(CodeRunner, TKSlot, boolean, int)

    /**
     * {@inheritDoc}
     */
    public void windowOpened(final WindowEvent we) {
        if (this.opening) {
            callback();
        } // end if
    } // end method 'windowOpened(WindowEvent)'

    /**
     * {@inheritDoc}
     */
    public void windowClosed(final WindowEvent we) {
        if (!this.opening) {
            callback();
        } // end if
    } // end method 'windowClosed(WindowEvent)'

    /**
     * Actually executes the callback.
     */
    private void callback() {
        try {
            this.codeRunner.callback(this.slot.getHandlerCode(),
                                     Value.createFromLong(id),
                                     Value.EMPTY_LIST);
        } catch (final Fail.Exception fe) {
            // nothing to do
        } catch (final Fatal.Exception fe) {
            // nothing to do
        } catch (final FalseExit fe) {
            // nothing to do
        } catch (final CadmiumException ce) {
            // nothing to do
        } // end try/catch
    } // end method 'callback()'

} // end class 'CamlWindowAdapter'
