/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;

import javax.swing.ImageIcon;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

import com.onemoonscientific.swank.ImageCmd;


/**
 * Implements all primitives from 'cltkImg.c'.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
@PrimitiveProvider
public final class CltkImg {

    /**
     * No instance of this class.
     */
    private CltkImg() {
    } // end empty constructor

    /**
     * Returns the pixels of a given image.
     * @param ctxt context
     * @param imgname image name
     * @return the pixels of the image, in byte-array RGB format
     * @throws Fail.Exception if no image exist with the given name
     */
    @Primitive
    public static Value camltk_getimgdata(final CodeRunner ctxt,
                                          final Value imgname)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        final BufferedImage bi = convertToBufferedImage(ImageCmd.getImage(imgname.asBlock().asString()));
        if (bi != null) {
            final int w = bi.getWidth();
            final int h = bi.getHeight();
            final byte[] bytes = new byte[w * h * 3];
            int ofs = 0;
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    final int rgb = bi.getRGB(x, y);
                    bytes[ofs++] = (byte) ((rgb >> 16) & 0xFF);
                    bytes[ofs++] = (byte) ((rgb >> 8) & 0xFF);
                    bytes[ofs++] = (byte) (rgb & 0xFF);
                } // end for
            } // end for
            final Block res = Block.createString(bytes);
            return Value.createFromBlock(res);
        } else {
            slot.tkError("no such image");
            return Value.UNIT; // never reached
        } // end if/else
    } // end method 'camltk_getimgdata(CodeRunner, Value)'

    /**
     * Modifies an image by drawing a rectangular array of pixels onto it.
     * @param ctxt context
     * @param imgname name of image to modify
     * @param pixmap pixels used to modify image, in byte-array RGB format
     * @param x where to put pixels in image
     * @param y where to put pixels in image
     * @param w width of rectangular array
     * @param h height of rectangular array
     * @return <i>unit</i>
     * @throws Fail.Exception if no image exist with the given name
     */
    @Primitive
    @SuppressWarnings("unchecked")
    public static Value camltk_setimgdata_native(final CodeRunner ctxt,
                                                 final Value imgname,
                                                 final Value pixmap,
                                                 final Value x,
                                                 final Value y,
                                                 final Value w,
                                                 final Value h)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        final Object original = ImageCmd.getImage(imgname.asBlock().asString());
        final BufferedImage bi = convertToBufferedImage(original);
        if (bi != null) {
            final byte[] bytes = pixmap.asBlock().getBytes();
            final int size = bytes.length / 3;
            int ofs = 0;
            for (int dy = 0; dy < h.asLong(); dy++) {
                for (int dx = 0; dx < w.asLong(); dx++) {
                    bi.setRGB(x.asLong() + dx,
                              y.asLong() + dy,
                              (bytes[ofs] << 16) | (bytes[ofs + 1] << 8) | bytes[ofs + 2]);
                    ofs += 3;
                } // end for
            } // end for
            if (original instanceof ImageIcon) {
                ((ImageIcon) original).setImage(bi);
            } else {
                ImageCmd.images.put(imgname.asBlock().asString(),
                                    new ImageIcon(bi));
            } // end if/else
            return Value.UNIT;
        } else {
            slot.tkError("no such image");
            return Value.UNIT; // never reached
        } // end if/else
    } // end method 'camltk_setimgdata_native(CodeRunner, Value, Value, Value, Value, Value, Value)'

    /**
     * Modifies an image by drawing a rectangular array of pixels onto it.
     * @param ctxt context
     * @param imgname name of image to modify
     * @param pixmap pixels used to modify image, in byte-array RGB format
     * @param x where to put pixels in image
     * @param y where to put pixels in image
     * @param w width of rectangular array
     * @param h height of rectangular array
     * @return <i>unit</i>
     * @throws Fail.Exception if no image exist with the given name
     */
    @Primitive
    public static Value camltk_setimgdata_bytecode(final CodeRunner ctxt,
                                                   final Value imgname,
                                                   final Value pixmap,
                                                   final Value x,
                                                   final Value y,
                                                   final Value w,
                                                   final Value h)
        throws Fail.Exception {
        return camltk_setimgdata_native(ctxt, imgname, pixmap, x, y, w, h);
    } // end method 'camltk_setimgdata_bytecode(CodeRunner, Value, Value, Value, Value, Value, Value)'

    /**
     * Converts an object into a buffered image.
     * @param obj object to convert
     * @return a buffered image if the passed object is a buffered image/bare image/image icon,
     *         <tt>null</tt> otherwise
     */
    private static BufferedImage convertToBufferedImage(final Object obj) {
        Image img;
        if (obj instanceof BufferedImage) {
            img = (BufferedImage) obj;
        } else if (obj instanceof Image) {
            img = (Image) obj;
        } else if (obj instanceof ImageIcon) {
            img = ((ImageIcon) obj).getImage();
        } else {
            return null;
        } // end if/elsif/else
        if (img instanceof BufferedImage) {
            return (BufferedImage) img;
        } // end if
        final int w = img.getWidth(null);
        final int h = img.getHeight(null);
        final BufferedImage res = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        final Graphics g = res.getGraphics();
        g.drawImage(img, 0, 0, null);
        g.dispose();
        return res;
    } // end method 'convertToBufferedImage(Object)'

} // end class 'CltkImg'
