/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.cadmium.Cadmium;

import tcl.lang.Interp;
import tcl.lang.TCL;
import tcl.lang.TclException;
import tcl.lang.TclInteger;
import tcl.lang.TclList;
import tcl.lang.TclObject;
import tcl.lang.TclString;

/**
 * Implements all primitives from 'cltkMain.c'.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
@PrimitiveProvider
public final class CltkMain {

    /**
     * No instance of this class.
     */
    private CltkMain() {
    } // end empty constructor

    /** Name of resource file for Tk. */
    private static final String RCNAME = ".camltkrc";

    /** Name of custom Tcl command used to execute callbacks. */
    private static final String CAMLCB = "camlcb";

    /**
     * Initializes the Tk machinery.
     * @param ctxt context
     * @param argv arguments for Tk initialization
     * @return <i>unit</i>
     * @throws Fail.Exception if <i>argv</i> is empty
     */
    @Primitive
    public static Value camltk_opentk(final CodeRunner ctxt, final Value argv)
        throws Fail.Exception {
        final Context c = ctxt.getContext();
        final TKSlot slot = (TKSlot) c.getSlot(TKSlot.SLOT);
        Interp interp = null;

        if (argv == Value.EMPTY_LIST) {
            Fail.failWith("camltk_opentk: argv is empty");
        } // end if
        String argv0 = argv.asBlock().get(0).asBlock().asString();

        try {
            interp = new Interp();
            slot.setInterpreter(interp);
            final Value interpCaml = ctxt.getContext().getCallback("cltclinterp");
            if (interpCaml != null) {
                interpCaml.asBlock().set(0, Cadmium.createObject(interp));
            } // end if
            interp.setVar("argv0", TclString.newInstance(argv0), TCL.GLOBAL_ONLY);
            Value tmp = argv.asBlock().get(1);
            int argc = 0;
            final TclObject tclArgv = TclList.newInstance();
            final List<String> args = new ArrayList<String>();
            while (tmp != Value.EMPTY_LIST) {
                final Block tmpBlock = tmp.asBlock();
                final String s = tmpBlock.get(0).asBlock().asString();
                args.add(s);
                TclList.append(interp, tclArgv, TclString.newInstance(s));
                tmp = tmpBlock.get(1);
                argc++;
            } // end while
            if (argc != 0) {
                interp.setVar("argv", tclArgv, TCL.GLOBAL_ONLY);
                interp.setVar("argc", TclInteger.newInstance(argc), TCL.GLOBAL_ONLY);
            } // end if

            Misc.evalResource(interp, "/com/onemoonscientific/swank/library/init.tcl");

            final int size = args.size();
            final int idxGeometry = args.indexOf("-geometry");
            if ((idxGeometry >= 0) && (idxGeometry < size - 1)) {
                Misc.eval(interp, "wm geometry . " + args.get(idxGeometry + 1));
            } // end if
            final int idxName = args.indexOf("-name");
            if ((idxName >= 0) && (idxName < size - 1)) {
                final StringBuilder sb = new StringBuilder();
                for (char ch : args.get(idxName + 1).toCharArray()) {
                    switch (ch) {
                    case '"':
                        sb.append("\\\"");
                        break;
                    case '\\':
                        sb.append("\\\\");
                        break;
                    default:
                        sb.append(ch);
                        break;
                    } // end switch
                } // end for
                Misc.eval(interp, "wm title . \"" + sb.toString() + "\"");
            } // end if
        } catch (final TclException te) {
            slot.tkError(te);
        } // end try/catch

        try {
            interp.createCommand(CltkMain.CAMLCB, new CamlCallbackCommand(ctxt, slot));
            interp.setVar("tcl_interactive", TclString.newInstance("0"), TCL.GLOBAL_ONLY);
            interp.setVar("BreakBindingsSequence", TclString.newInstance("0"), TCL.GLOBAL_ONLY);
        } catch (final TclException te) {
            slot.tkError(te);
        } // end try/catch

        // load rc file
        try {
            final String home = System.getenv("HOME");
            if (home != null) {
                final String file = home + "/" + CltkMain.RCNAME;
                if ((new File(file)).canRead()) {
                    try {
                        Misc.evalFile(interp, file);
                    } catch (final TclException te) {
                        slot.tkError(te);
                    } // end try/catch
                } // end if
            } // end if
        } catch (final SecurityException se) {
            // nothing to do, just ignore file
        } // end try/catch
        return Value.UNIT;
    } // end method 'camltk_opentk(CodeRunner, Value)'

    /**
     * Disposes the Tcl interpreter.
     * @param ctxt context
     * @param unit ignored
     * @return <i>unit</i>
     */
    @Primitive
    public static Value camltk_finalize(final CodeRunner ctxt, final Value unit)
        throws Fail.Exception {
        try {
            Misc.disposeInterp(TKSlot.checkInit(ctxt).getInterpreter());
        } catch (final Throwable t) {
            // fail silently
        } // end try/catch
        return Value.UNIT;
    } // end method 'camltk_finalize(CodeRunner, Value)'

} // end class 'CltkMain'
