/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

import tcl.lang.TclException;
import tcl.lang.TclObject;
import tcl.lang.TclString;
import tcl.lang.TCL;

/**
 * Implements all primitives from 'cltkVar.c'.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
@PrimitiveProvider
public final class CltkVar {

    /**
     * No instance of this class.
     */
    private CltkVar() {
    } // end empty constructor

    /**
     * Retrieves the value of a variable.
     * @param ctxt context
     * @param var variable name
     * @return the value of the variable whose name is passed
     * @throws Fail.Exception if no variable exists with the passed name
     */
    @Primitive
    public static Value camltk_getvar(final CodeRunner ctxt,
                                      final Value var)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        try {
            final TclObject res = slot.getInterpreter().getVar(var.asBlock().asString(),
                                                               TCL.GLOBAL_ONLY);
            return Value.createFromBlock(Block.createString(res.toString()));
        } catch (final TclException te) {
            slot.tkError(te);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'camltk_getvar(CodeRunner, Value)'

    /**
     * Changes the value of a variable.
     * @param ctxt context
     * @param var name of variable to modify
     * @param contents new value for variable
     * @return <i>unit</i>
     * @throws Fail.Exception if no variable exists with the passed name
     */
    @Primitive
    public static Value camltk_setvar(final CodeRunner ctxt,
                                      final Value var,
                                      final Value contents)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        try {
            final TclObject cont = TclString.newInstance(contents.asBlock().asString());
            slot.getInterpreter().setVar(var.asBlock().asString(),
                                         cont,
                                         TCL.GLOBAL_ONLY);
        } catch (final TclException te) {
            slot.tkError(te);
        } // end try/catch
        return Value.UNIT;
    } // end method 'camltk_setvar(CodeRunner, Value, Value)'

    /**
     * Installs a handler tracing variable modifications.
     * @param ctxt context
     * @param var name of variable to trace
     * @param cbid identifier of callback to invoke on variable modification
     * @return <i>unit</i>
     * @throws Fail.Exception if no variable exists with the passed name
     */
    @Primitive
    public static Value camltk_trace_var(final CodeRunner ctxt,
                                         final Value var,
                                         final Value cbid)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        try {
            slot.getInterpreter().traceVar(var.asBlock().asString(),
                                           slot.createVarTrace(ctxt, cbid.asLong()),
                                           TCL.GLOBAL_ONLY
                                           | TCL.TRACE_WRITES
                                           | TCL.TRACE_UNSETS);
        } catch (final TclException te) {
            slot.tkError(te);
        } // end try/catch
        return Value.UNIT;
    } // end method 'camltk_trace_var(CodeRunner, Value, Value)'

    /**
     * Uninstalls a handler tracing variable modifications.
     * @param ctxt context
     * @param var name of variable to cease trace for
     * @param cbid identifier of callback registered to be invoked on variable modification
     * @return <i>unit</i>
     */
    @Primitive
    public static Value camltk_untrace_var(final CodeRunner ctxt,
                                           final Value var,
                                           final Value cbid)
        throws Fail.Exception {
        final TKSlot slot = TKSlot.checkInit(ctxt);
        slot.getInterpreter().untraceVar(var.asBlock().asString(),
                                         slot.getVarTrace(cbid.asLong()),
                                         TCL.GLOBAL_ONLY
                                         | TCL.TRACE_WRITES
                                         | TCL.TRACE_UNSETS);
        return Value.UNIT;
    } // end method 'camltk_untrace_var(CodeRunner, Value, Value)'

} // end class 'CltkVar'
