/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.labltk;

import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.Value;

import tcl.lang.Interp;
import tcl.lang.TimerHandler;
import tcl.lang.VarTrace;

/**
 * This class represents the context associated with the 'labltk' library.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public final class TKSlot {

    /** Identifier for <tt>TKSlot</tt>. */
    public static final Object SLOT = new Object();

    /** Value associated with the <tt>Protocol.TkError</tt> exception. */
    private final Value tkError;

    /** Value associated with the <tt>Protocol.protected_dispatch</tt> function. */
    private final Value handlerCode;

    /** Interpreter associated with slot. */
    private Interp interpreter;

    /** Map from identifiers to variable tracers. */
    private final Map<Integer, VarTrace> varTraces;

    /** Map from identifiers to timer handlers. */
    private final Map<Integer, TimerHandler> timerHandlers;

    /** Map from identifiers to waiter threads. */
    private final Map<Long, InputWaiterThread> inputWaiterThreads;

    /**
     * Constructs a slot from a context, retrieving callbacks.
     * @param ctxt context used to retrieve callbacks - should not be <tt>null</tt>
     */
    public TKSlot(final Context ctxt) {
        assert ctxt != null : "null ctxt";
        this.tkError = ctxt.getCallback("tkerror");
        this.handlerCode = ctxt.getCallback("camlcb");
        this.interpreter = null;
        this.varTraces = new HashMap<Integer, VarTrace>();
        this.timerHandlers = new HashMap<Integer, TimerHandler>();
        this.inputWaiterThreads = new HashMap<Long, InputWaiterThread>();
    } // end constructor(Context)

    /**
     * Raises the <tt>Protocol.TkError</tt> exception with a given message.
     * @param msg message used as parameter for exception - should not be <tt>null</tt>
     * @throws Fail.Exception always
     */    
    public void tkError(final String msg) throws Fail.Exception {
        assert msg != null : "null msg";
        Fail.raiseWithString(this.tkError, msg);
    } // end method 'tkError(String)'

    /**
     * Raises the <tt>Protocol.TkError</tt> exception with a given message.
     * @param t exception to get message from - should not be <tt>null</tt>
     * @throws Fail.Exception always
     */    
    public void tkError(final Throwable t) throws Fail.Exception {
        assert t != null : "null t";
        Fail.raiseWithString(this.tkError, t.getMessage());
    } // end method 'tkError(Throwable)'

    /**
     * Returns the interpreter associated with the slot.
     * @return the interpreter associated with the slot
     */
    public Interp getInterpreter() {
        return this.interpreter;
    } // end method 'getInterpreter()'

    /**
     * Changes the interpreter associated with the slot.
     * @param i new interpreter - should not be <tt>null</tt>
     */
    public void setInterpreter(final Interp i) {
        assert i != null : "null i";
        this.interpreter = i;
    } // end method 'setInterpreter(Interp)'

    /**
     * Returns the value associated with the <tt>Protocol.protected_dispatch</tt> function.
     * @return the value associated with the <tt>Protocol.protected_dispatch</tt> function
     */
    public Value getHandlerCode() {
        return this.handlerCode;
    } // end method 'getHandlerCode()'

    /**
     * Returns the variable tracer for a given identifier.
     * @param id tracer identifier
     * @return the variable tracer for the passed identifier
     */
    public VarTrace getVarTrace(final int id) {
        return this.varTraces.get(id);
    } // end method 'getVarTrace(int)'

    /**
     * Creates a variable tracer.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param id callback identifier
     * @return a variable tracer calling passed callback
     */
    public VarTrace createVarTrace(final CodeRunner cr, final int id) {
        final VarTrace res = new CamlVarTrace(cr, this, id);
        this.varTraces.put(id, res);
        return res;
    } // end method 'createVarTrace(CodeRunner, int)'

    /**
     * Returns the timer handler for a given identifier.
     * @param id handler identifier
     * @return the timer handler for the passed identifier
     */
    public TimerHandler getTimerHandler(final int id) {
        return this.timerHandlers.get(id);
    } // end method 'getTimerHandler(int)'

    /**
     * Creates a timer handler.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param milli timer duration in milliseconds
     * @param cb callback identifier
     * @return the identifier associated with the created handler
     */
    public int createTimerHandler(final CodeRunner cr, final int milli, final int cb) {
        final int res = this.timerHandlers.size();
        final TimerHandler th = new CamlTimer(cr, this, cb, milli);
        this.timerHandlers.put(res, th);
        return res;
    } // end method 'createTimerHandler(CodeRunner, int, int)'

    /**
     * Creates and starts an input waiter thread.
     * @param cr used to execute callback - should not be <tt>null</tt>
     * @param fd file descriptor
     * @param cb callback
     */
    public void createInputWaiterThread(final CodeRunner cr, final int fd, final int cb) {
        final long id = ((long) fd) << 32 | ((long) cb);
        final Channel ch = cr.getContext().getChannel(fd);
        final InputStream st = ch != null ? ch.asInputStream() : null;
        if (st != null) {
            final InputWaiterThread iwt = new InputWaiterThread(cr, this, cb, st);
            this.inputWaiterThreads.put(id, iwt);
            iwt.start();
        } // end if
    } // end method 'createInputWaiterThread(CodeRunner, int, int)'

    /**
     * Cancels an input waiter thread.
     * @param fd file descriptor
     * @param cb callback
     */
    public void removeInputWaiterThread(final int fd, final int cb) {
        final long id = ((long) fd) << 32 | ((long) cb);
        final InputWaiterThread iwt = this.inputWaiterThreads.get(id);
        if (iwt != null) {
            iwt.interrupt();
        } // end if
    } // end method 'removeInputWaiterThread(int, int)'

    /**
     * Checks that the library is fully initialized.
     * @param ctxt context to check initialization for - should not be <tt>null</tt>
     * @return the context associated with the library
     * @throws Fail.Exception if the library is not fully initialized in the passed context
     */
    static TKSlot checkInit(final CodeRunner ctxt) throws Fail.Exception {
        assert ctxt != null : "null ctxt";
        final Context c = ctxt.getContext();
        final TKSlot slot = (TKSlot) c.getSlot(SLOT);
        if (slot != null) {
            if (slot.getInterpreter() == null) {
                slot.tkError("Tcl/Tk not initialised");
            } // end if
        } else {
            Fail.invalidArgument("Tcl/Tk not initialised");
        } // end if/else
        return slot;
    } // end method 'checkInit(CodeRunner)'

} // end class 'TKSlot'
