/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.unix;

import java.io.File;

import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.util.Misc;

/**
 * This class provides implementation for 'unix_chmod' primitive.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Chmod {

    /**
     * No instance of this class.
     */
    private Chmod() {
    } // end empty constructor

    /**
     * Tries to change file permissions by calling 'chmod' executable.
     * @param ctxt context
     * @param path file to set permissions for
     * @param perm permissions for file
     * @return <i>unit</i>
     * @throws Fail.Exception if permissions change fails
     */
    @Primitive
    public static Value unix_chmod(final CodeRunner ctxt,
                                   final Value path,
                                   final Value perm)
        throws Fail.Exception, FalseExit {
        final File f = ctxt.getContext().getRealFile(path);
        if (ctxt.getContext().getParameters().isUnixEmulated()) {
            final String res = Unix.emul(ctxt,
                                         "chmod",
                                         "chmod",
                                         Integer.toString(perm.asLong(), 8),
                                         f.getAbsolutePath());
            if ((res == null) || (res.length() != 0)) {
                Unix.fail(ctxt, "chmod", "unable to change file permissions");
            } // end if
        } else {
            try {
                Misc.setPermissions(f, perm.asLong());
            } catch (final Exception e) {
                Unix.fail(ctxt, "chmod", "unable to change file permissions");
            } // end try/catch
        } // end if/else
        return Value.UNIT;
    } // end method 'unix_chmod(CodeRunner, Value, Value)'

} // end class 'Chmod'
