/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.unix;

import java.io.IOException;
import java.io.InterruptedIOException;

import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Misc;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Sys;
import fr.x9c.cadmium.util.StreamCopyThread;

/**
 * This class provides implementation for some execution-related primitives.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Execv {

    /**
     * No instance of this class.
     */
    private Execv() {
    } // end empty constructor

    /**
     * Simulates <i>execv</i> by executing given a program, then stopping
     * program.
     * @param ctxt context
     * @param path program to execute
     * @param args program arguments
     * @return <i>unit</i>
     * @see fr.x9c.cadmium.primitives.stdlib.Sys#caml_sys_exit(CodeRunner, Value)
     * @throws Fail.Exception if program cannot be executed
     */
    @Primitive
    public static Value unix_execv(final CodeRunner ctxt,
                                   final Value path,
                                   final Value args)
        throws Fail.Exception, Fatal.Exception, FalseExit {
        final Context context = ctxt.getContext();
        final int sz = args.asBlock().getWoSize();
        final String[] a = new String[sz];
        for (int i = 0; i < sz; i++) {
            a[i] = args.asBlock().get(i).asBlock().asString();
        } // end for
        a[0] = context.getRealFile(path).getAbsolutePath();
        try {
            final ProcessBuilder pb =
                new ProcessBuilder(Misc.prepareArguments(a));
            pb.directory(context.getPwd());
            final Process p = pb.start();
            final Channel out = context.getChannel(Channel.STDOUT);
            if (out != null) {
                final Thread t = new StreamCopyThread(out.asOutputStream(),
                                                      p.getInputStream());
                t.start();
            } // end if
            final Channel err = context.getChannel(Channel.STDERR);
            if (err != null) {
                final Thread t = new StreamCopyThread(err.asOutputStream(),
                                                      p.getErrorStream());
                t.start();
            } // end if
            final Channel in = context.getChannel(Channel.STDIN);
            if (in != null) {
                final Thread t = new StreamCopyThread(p.getOutputStream(),
                                                      in.asInputStream());
                t.start();
            } // end if
            return Sys.caml_sys_exit(ctxt, Value.createFromLong(p.waitFor()));
        } catch (final FalseExit sfe) {
            throw sfe;
        } catch (final InterruptedException ie) {
            Unix.fail(ctxt, "execv", ie);
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe = FalseExit.createFromContext(context);
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Unix.fail(ctxt, "execv", ioe);
        } // end try/catch
        return Value.UNIT;
    } // end method 'unix_execv(CodeRunner, Value, Value)'

} // end class 'Execv'
