/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.unix;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Misc;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides implementation for 'unix_open' primitive.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Open {

    /** For convesion of open flags. */
    private static final int[] OPEN_FLAGS = {
        Channel.O_RDONLY,
        Channel.O_WRONLY,
        Channel.O_RDONLY | Channel.O_WRONLY,
        Channel.O_NONBLOCK,
        Channel.O_APPEND,
        Channel.O_CREAT,
        Channel.O_TRUNC,
        Channel.O_EXCL,
        0,
        0,
        0,
        0
    };

    /**
     * No instance of this class.
     */
    private Open() {
    } // end empty constructor

    /**
     * Opens a file.
     * @param ctxt context
     * @param path file to open
     * @param flags access control
     * @param perm file permissions if file is created
     * @return file descriptor
     * @throws Fail.Exception if an i/o error occurs
     */
    @Primitive
    public static Value unix_open(final CodeRunner ctxt,
                                  final Value path,
                                  final Value flags,
                                  final Value perm)
        throws Fail.Exception, FalseExit {
        // cannot rely on caml_sys_open(-) as flags are not the same
        final Context context = ctxt.getContext();
        final File file = context.getRealFile(path);
        final int flg = Misc.convertFlagList(flags, Open.OPEN_FLAGS);
        final int perms = perm.asLong();
        try {
            context.enterBlockingSection();
            final Value res =
                Value.createFromLong(context.addChannel(new Channel(ctxt,
                                                                    file,
                                                                    flg,
                                                                    perms)));
            context.leaveBlockingSection();
            return res;
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe = FalseExit.createFromContext(context);
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            context.leaveBlockingSection();
            Unix.fail(ctxt, "open", ioe);
            return Value.UNIT; // never reached
        } // end try/catch
    } // end method 'unix_open(CodeRunner, Value)'

} // end class 'Open'
