/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.primitives.unix;

import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.RandomAccessFile;

import fr.x9c.cadmium.kernel.CodeRunner;
import fr.x9c.cadmium.kernel.Fail;
import fr.x9c.cadmium.kernel.FalseExit;
import fr.x9c.cadmium.kernel.Primitive;
import fr.x9c.cadmium.kernel.PrimitiveProvider;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class provides implementation for truncate-related primitives.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
@PrimitiveProvider
public final class Truncate {

    /**
     * No instance of this class.
     */
    private Truncate() {
    } // end empty constructor

    /**
     * Changes the length of a file.
     * @param ctxt context
     * @param path file to modify
     * @param len new file length
     * @return <i>unit</i>
     * @throws Fail.Exception if file cannot be truncated
     */
    @Primitive
    public static Value unix_truncate(final CodeRunner ctxt,
                                      final Value path,
                                      final Value len)
        throws Fail.Exception, FalseExit {
        try {
            final File f = ctxt.getContext().getRealFile(path);
            (new RandomAccessFile(f, "rw")).setLength(len.asLong());
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Unix.fail(ctxt, "truncate", ioe);
        } // end try/catch
        return Value.UNIT;
    } // end method 'unix_truncate(CodeRunner, Value, Value)'

    /**
     * Changes the length of a file.
     * @param ctxt context
     * @param path file to modify
     * @param len new file length
     * @return <i>unit</i>
     * @throws Fail.Exception if file cannot be truncated
     */
    @Primitive
    public static Value unix_truncate_64(final CodeRunner ctxt,
                                         final Value path,
                                         final Value len)
        throws Fail.Exception, FalseExit {
        try {
            final File f = ctxt.getContext().getRealFile(path);
            (new RandomAccessFile(f, "rw")).setLength(len.asBlock().asInt64());
        } catch (final InterruptedIOException iioe) {
            final FalseExit fe =
                FalseExit.createFromContext(ctxt.getContext());
            fe.fillInStackTrace();
            throw fe;
        } catch (final IOException ioe) {
            Unix.fail(ctxt, "truncate_64", ioe);
        } // end try/catch
        return Value.UNIT;
    } // end method 'unix_truncate_64(CodeRunner, Value, Value)'

} // end class 'Truncate'
