/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.support.servlet;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import javax.servlet.ServletException;

import fr.x9c.cadmium.kernel.ByteCodeParameters;
import fr.x9c.cadmium.kernel.CadmiumException;
import fr.x9c.cadmium.kernel.Context;
import fr.x9c.cadmium.kernel.Fatal;
import fr.x9c.cadmium.kernel.Interpreter;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.util.RandomAccessInputStream;

/**
 * This class implements a servlet whose behaviour is defined by callbacks
 * registered on the OCaml side using <tt>Servlet.register</tt>.
 * An interpreter is created by the constructor of the class, and this
 * interpreter is subsequently used to run the callbacks. <br/>
 * <br/>
 * Bytecode is loaded from the resource <tt>/resources/cadmium/servlet</tt>
 * and run using parameters loaded from
 * <tt>/resources/cadmium/parameters.properties</tt>. <br/>
 * <br/>
 * One may use the <tt>cadmium.make-servlet</tt> ant macro to easily create
 * a war file with the requested setting.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.1
 */
public final class ByteCodeServlet extends Servlet {

    /** Serialization UID. */
    static final long serialVersionUID = -2660681303415309297L;

    /** Interpreter of servlet code. */
    static Interpreter interpreter;

    /** Path to 'servlet parameters' resource. */
    private static final String PARAMETERS = "/resources/cadmium/parameters.properties";

    /** Path to 'servlet bytecode' resource. */
    private static final String SERVLET = "/resources/cadmium/servlet";

    /**
     * Constructs a servlet runner from resources.
     * @throws IOException if an i/o error occurs
     * @throws Fatal.Exception if bytecode or parameters could not be loaded
     * @throws CadmiumException if interpreter could not be created
     */
    public ByteCodeServlet()
        throws IOException, Fatal.Exception, CadmiumException {
        super();

        // resources
        final InputStream propsStream =
            ByteCodeServlet.class.getResourceAsStream(PARAMETERS);
        final InputStream bytecodeStream =
            ByteCodeServlet.class.getResourceAsStream(SERVLET);
        if ((propsStream == null) || (bytecodeStream == null)) {
            Fatal.raise("bytecode/parameters ressource is missing");
        } // end if

        // interpreter parameters
        final Properties props = new Properties();
        props.load(propsStream);
        final ByteCodeParameters params =
            ByteCodeParameters.fromProperties(props,
                                              "servlet",
                                              new String[0],
                                              System.in,
                                              System.out,
                                              System.err);
        propsStream.close();

        // interpreter execution
        final RandomAccessInputStream bytecode =
            new RandomAccessInputStream(bytecodeStream);

        final String[] oldProviders = params.getProviders();
        final int len = oldProviders.length;
        final String[] newProviders = new String[len + 1];
        System.arraycopy(oldProviders, 0, newProviders, 0, len);
        newProviders[len] = "fr.x9c.cadmium.primitives.servlet.Servlets";

        ByteCodeServlet.interpreter =
            new Interpreter(new ByteCodeParameters(params.getArguments(),
                                                   false, /* backtrace */
                                                   false, /* stop JVM */
                                                   params.getStandardInput(),
                                                   params.getStandardOutput(),
                                                   params.getStandardError(),
                                                   true, /* awt */
                                                   true, /* bare canvas */
                                                   params.isJavaxSoundUsed(),
                                                   params.isJDBMUsed(),
                                                   "Unix", /* OS */
                                                   params.isUnixEmulated(),
                                                   params.getFile(),
                                                   false, /* embedded */
                                                   null,  /* embedded base */
                                                   false, /* dialog */
                                                   false, /* help */
                                                   false, /* version */
                                                   false, /* disassembled */
                                                   params.getInitStackSize(),
                                                   params.getMaxStackSize(),
                                                   newProviders,
                                                   params.isDispatcherCompiled()),
                            new File("."),
                            bytecode);
        bytecode.close();
        ByteCodeServlet.interpreter.execute();
    } // end empty constructor

    /**
     * {@inheritDoc}
     */
    protected final Value callback(final String cb, final Value... args)
        throws ServletException {
        assert cb != null : "null cb";
        try {
            return ByteCodeServlet.interpreter.execute(cb, args);
        } catch (final Exception e) {
            throw new ServletException("Callback error", e);
        } // end try/catch
    } // end method 'callback(String, Value...)'

    /**
     * {@inheritDoc}
     */
    protected final void callbackNoException(final String cb,
                                             final Value... args) {
        assert cb != null : "null cb";
        try {
            ByteCodeServlet.interpreter.execute(cb, args);
        } catch (final Exception e) {
            // nothing to do
        } // end try/catch
    } // end method 'callback(String, Value...)'

    /**
     * {@inheritDoc}
     */
    protected final boolean callbackOption(final String cb,
                                           final Value... args)
        throws ServletException {
        assert cb != null : "null cb";
        if (getCallback(cb) != null) {
            try {
                ByteCodeServlet.interpreter.execute(cb, args);
            } catch (final Exception e) {
                throw new ServletException("Callback error", e);
            } // end try/catch
            return false;
        } else {
            return true;
        } // end if/else
    } // end method 'callbackOption(String, Value...)'

    /**
     * {@inheritDoc}
     */
    protected final Value getCallback(final String cb) {
        return ByteCodeServlet.interpreter.getContext().getCallback(cb);
    } // end method 'getCallback(String)'

    /**
     * {@inheritDoc}
     */
    protected final Context getContext() {
        return ByteCodeServlet.interpreter.getContext();
    } // end method 'getContext()'

} // end class 'ByteCodeServlet'
