/*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.cadmium.support.values;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class defines values of the <i>string</i> type.
 *
 * @author <a href="mailto:cadmium@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public final class MutableString implements ToValue {

    /** The actual value. */
    private final StringBuilder value;

    /**
     * Constructs a string value.
     * @param v string value - should not be <tt>null</tt>
     */
    public MutableString(final String v) {
        assert v != null : "null v";
        this.value = new StringBuilder(v);
    } // end constructor(String)

    /**
     * Returns the value, as a string.
     * @return the value
     */
    public String get() {
        return this.value.toString();
    } // end method 'get()'

    /**
     * Returns the character at a given position.
     * @param index character position
     * @return the character at the passed position
     */
    public char getChar(final int index) {
        return this.value.charAt(index);
    } // end method 'getChar(int)'

    /**
     * Changes a character.
     * @param index position of character to change
     * @param v new character value
     */
    public void setChar(final int index, final char v) {
        this.value.setCharAt(index, v);
    } // end method 'setChar(int, char)'

    /**
     * {@inheritDoc}
     */
    @Override
    public Value toValue() {
        return Value.createFromBlock(Block.createString(this.value.toString()));
    } // end method 'toValue()'

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return this.value.hashCode();
    } // end method 'hashCode()'

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof MutableString) {
            final MutableString that = (MutableString) obj;
            return this.value.equals(that.value);
        } else {
            return false;
        } // end if/else
    } // end method 'equals(Object)'

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return this.value.toString();
    } // end method 'toString()'

} // end class 'MutableString'
