(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Types *)

type java_object

type java_class = java_object

type java_constructor = java_object

type java_field = java_object

type java_method = java_object

type 'c java_type =
  | Boolean
  | Byte
  | Char
  | Double
  | Float
  | Int
  | Long
  | Short
  | String
  | Class of 'c
  | Void
  | Array of 'c java_type

type 'o java_value =
  | Boolean_val of bool
  | Byte_val of int
  | Char_val of int
  | Double_val of float
  | Float_val of float
  | Int_val of int32
  | Long_val of int64
  | Short_val of int
  | String_val of string
  | Object_val of 'o

type java_modifier =
  | Abstract
  | Final
  | Interface
  | Native
  | Private
  | Protected
  | Public
  | Static
  | StrictFP
  | Synchronized
  | Transcient
  | Volatile


(* Base functions *)

external is_available : unit -> bool =
  "cadmium_is_available"

external is_null : java_object -> bool =
  "cadmium_is_null"

external get_null : unit -> java_object =
  "cadmium_get_null"

external instance_of : java_object -> java_class -> bool =
  "cadmium_instance_of"

external cast : java_class -> java_object -> java_object =
  "cadmium_class_cast"

external java_string_of_string : string -> java_object =
  "cadmium_object_of_string"

external string_of_java_string : java_object -> string =
  "cadmium_object_to_string"

external java_char_of_char : char -> int =
  "cadmium_java_char_of_char"

external char_of_java_char : int -> char =
  "cadmium_char_of_java_char"

external equal : java_object -> java_object -> bool =
  "cadmium_equal"

external not_equal : java_object -> java_object -> bool =
  "cadmium_not_equal"


(* java.lang.Object primitives *)

module Object =
  struct
    external make : unit -> java_object =
      "cadmium_object_make"

    external clone : java_object -> java_object =
      "cadmium_object_clone"

    external equals : java_object -> java_object -> bool =
      "cadmium_object_equals"

    external get_class : java_object -> java_class =
      "cadmium_object_get_class"

    external hash_code : java_object -> int32 =
      "cadmium_object_hash_code"

    external notify : java_object -> unit =
      "cadmium_object_notify"

    external notify_all : java_object -> unit =
      "cadmium_object_notify_all"

    external to_string : java_object -> string =
      "cadmium_object_to_string"

    external wait : java_object -> unit =
      "cadmium_object_wait"

    external wait_time : java_object -> int64 -> unit =
      "cadmium_object_wait_time1"

    external wait_time' : java_object -> int64 -> int32 -> unit =
      "cadmium_object_wait_time2"
  end


(* java.lang.Class primitives *)

module Class =
  struct
    external cast : java_class -> java_object -> java_object =
      "cadmium_class_cast"

    external for_name : string -> java_class =
      "cadmium_class_for_name"

    external get_canonical_name : java_class -> string =
      "cadmium_class_get_canonical_name"

    external get_constructor : java_class -> java_class java_type list -> java_constructor =
      "cadmium_class_get_constructor"

    external get_constructors : java_class -> java_constructor list =
      "cadmium_class_get_constructors"

    external get_declared_constructor : java_class -> java_class java_type list -> java_constructor =
      "cadmium_class_get_declared_constructor"

    external get_declared_constructors : java_class -> java_constructor list =
      "cadmium_class_get_declared_constructors"

    external get_field : java_class -> string -> java_field =
      "cadmium_class_get_field"

    external get_fields : java_class -> java_field list =
      "cadmium_class_get_fields"

    external get_declared_field : java_class -> string -> java_field =
      "cadmium_class_get_declared_field"

    external get_declared_fields : java_class -> java_field list =
      "cadmium_class_get_declared_fields"

    external get_method : java_class -> string -> java_class java_type list -> java_method =
      "cadmium_class_get_method"

    external get_methods : java_class -> java_method list =
      "cadmium_class_get_methods"

    external get_declared_method : java_class -> string -> java_class java_type list -> java_method =
      "cadmium_class_get_declared_method"

    external get_declared_methods : java_class -> java_method list =
      "cadmium_class_get_declared_methods"

    external get_interfaces : java_class -> java_class list =
      "cadmium_class_get_interfaces"

    external get_modifiers : java_class -> java_modifier list =
      "cadmium_class_get_modifiers"

    external get_name : java_class -> string =
      "cadmium_class_get_name"

    external get_package : java_class -> string =
      "cadmium_class_get_package"

    external get_simple_name : java_class -> string =
      "cadmium_class_get_simple_name"

    external get_superclass : java_class -> java_class option =
      "cadmium_class_get_superclass"

    external is_array : java_class -> bool =
      "cadmium_class_is_array"

    external is_interface : java_class -> bool =
      "cadmium_class_is_interface"

    external is_primitive : java_class -> bool =
      "cadmium_class_is_primitive"

    external new_instance : java_class -> java_object =
      "cadmium_class_new_instance"

    let to_object obj = obj

    external as_class : java_object -> java_class =
      "cadmium_class_as_class"
  end


(* java.lang.reflect.Constructor primitives *)

module Constructor =
  struct
    external get_declaring_class : java_constructor -> java_class =
      "cadmium_constr_get_declaring_class"

    external get_exception_types : java_constructor -> java_class list =
      "cadmium_constr_get_exception_types"

    external get_modifiers : java_constructor -> java_modifier list =
      "cadmium_constr_get_modifiers"

    external get_parameter_types : java_constructor -> java_class java_type list =
      "cadmium_constr_get_parameter_types"

    external new_instance : java_constructor -> java_object java_value list -> java_object =
      "cadmium_constr_new_instance"

    let to_object obj = obj

    external as_constructor : java_object -> java_constructor =
      "cadmium_constr_as_constr"
  end


(* java.lang.reflect.Field primitives *)

module Field =
  struct
    external get_declaring_class : java_field -> java_class =
      "cadmium_field_get_declaring_class"

    external get_modifiers : java_field -> java_modifier list =
      "cadmium_field_get_modifiers"

    external get_name : java_field -> string =
      "cadmium_field_get_name"

    external get : java_field -> java_object -> java_object =
      "cadmium_field_get"

    external get_boolean : java_field -> java_object -> bool =
      "cadmium_field_get_boolean"

    external get_byte : java_field -> java_object -> int =
      "cadmium_field_get_byte"

    external get_char : java_field -> java_object -> int =
      "cadmium_field_get_char"

    external get_double : java_field -> java_object -> float =
      "cadmium_field_get_double"

    external get_float : java_field -> java_object -> float =
      "cadmium_field_get_float"

    external get_int : java_field -> java_object -> int32 =
      "cadmium_field_get_int"

    external get_long : java_field -> java_object -> int64 =
      "cadmium_field_get_long"

    external get_short : java_field -> java_object -> int =
      "cadmium_field_get_short"

    external set : java_field -> java_object -> java_object -> unit =
      "cadmium_field_set"

    external set_boolean : java_field -> java_object -> bool -> unit =
      "cadmium_field_set_boolean"

    external set_byte : java_field -> java_object -> int -> unit =
      "cadmium_field_set_byte"

    external set_char : java_field -> java_object -> int -> unit =
      "cadmium_field_set_char"

    external set_double : java_field -> java_object -> float -> unit =
      "cadmium_field_set_double"

    external set_float : java_field -> java_object -> float -> unit =
      "cadmium_field_set_float"

    external set_int : java_field -> java_object -> int32 -> unit =
      "cadmium_field_set_int"

    external set_long : java_field -> java_object -> int64 -> unit =
      "cadmium_field_set_long"

    external set_short : java_field -> java_object -> int -> unit =
      "cadmium_field_set_short"

    let to_object obj = obj

    external as_field : java_object -> java_field =
      "cadmium_field_as_field"
  end


(* java.lang.reflect.Method primitives *)

module Method =
  struct
    external get_declaring_class : java_method -> java_class =
      "cadmium_meth_get_declaring_class"

    external get_exception_types : java_method -> java_class list =
      "cadmium_meth_get_exception_types"

    external get_modifiers : java_method -> java_modifier list =
      "cadmium_meth_get_modifiers"

    external get_name : java_method -> string =
      "cadmium_meth_get_name"

    external get_parameter_types : java_method -> java_class java_type list =
      "cadmium_meth_get_parameter_types"

    external get_return_type : java_method -> java_class java_type =
      "cadmium_meth_get_return_type"

    external invoke : java_method -> java_object -> java_object java_value list -> java_object java_value =
      "cadmium_meth_invoke"

    let to_object obj = obj

    external as_method : java_object -> java_method =
      "cadmium_meth_as_meth"
  end


(* java.lang.reflect.Array primitives *)

module Array =
  struct
    external new_instance : java_class java_type -> int -> java_object =
      "cadmium_array_new_instance"

    external new_instance_dims : java_class java_type -> int list -> java_object =
      "cadmium_array_new_instance_dims"

    external get_length : java_object -> int =
      "cadmium_array_get_length"

    external get : java_object -> int -> java_object =
      "cadmium_array_get"

    external get_boolean : java_object -> int -> bool =
      "cadmium_array_get_boolean"

    external get_byte : java_object -> int -> int =
      "cadmium_array_get_byte"

    external get_char : java_object -> int -> int =
      "cadmium_array_get_char"

    external get_double : java_object -> int -> float =
    "cadmium_array_get_double"

    external get_float : java_object -> int -> float =
      "cadmium_array_get_float"

    external get_int : java_object -> int -> int32 =
      "cadmium_array_get_int"

    external get_long : java_object -> int -> int64 =
      "cadmium_array_get_long"

    external get_short : java_object -> int -> int =
      "cadmium_array_get_short"

    external set : java_object -> int -> java_object -> unit =
      "cadmium_array_set"

    external set_boolean : java_object -> int -> bool -> unit =
      "cadmium_array_set_boolean"

    external set_byte : java_object -> int -> int -> unit =
      "cadmium_array_set_byte"

    external set_char : java_object -> int -> int -> unit =
      "cadmium_array_set_char"

    external set_double : java_object -> int -> float -> unit =
      "cadmium_array_set_double"

    external set_float : java_object -> int -> float -> unit =
      "cadmium_array_set_float"

    external set_int : java_object -> int -> int32 -> unit =
      "cadmium_array_set_int"

    external set_long : java_object -> int -> int64 -> unit =
      "cadmium_array_set_long"

    external set_short : java_object -> int -> int -> unit =
      "cadmium_array_set_short"
  end


(* Conversions between OCaml and Java arrays. *)

module ArrayUtils =
  struct
    let bool_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Boolean len in
      for i = 0 to len - 1 do
        Array.set_boolean res i (StdLabels.Array.get a i)
      done;
      res

    let byte_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Byte len in
      for i = 0 to len - 1 do
        Array.set_byte res i (StdLabels.Array.get a i)
      done;
      res

    let char_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Char len in
      for i = 0 to len - 1 do
        Array.set_char res i (java_char_of_char (StdLabels.Array.get a i))
      done;
      res

    let double_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Double len in
      for i = 0 to len - 1 do
        Array.set_double res i (StdLabels.Array.get a i)
      done;
      res

    let float_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Float len in
      for i = 0 to len - 1 do
        Array.set_float res i (StdLabels.Array.get a i)
      done;
      res

    let int_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Int len in
      for i = 0 to len - 1 do
        Array.set_int res i (StdLabels.Array.get a i)
      done;
      res

    let long_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Long len in
      for i = 0 to len - 1 do
        Array.set_long res i (StdLabels.Array.get a i)
      done;
      res

    let short_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance Short len in
      for i = 0 to len - 1 do
        Array.set_short res i (StdLabels.Array.get a i)
      done;
      res

    let string_array_to_java a =
      let len = StdLabels.Array.length a in
      let res = Array.new_instance String len in
      for i = 0 to len - 1 do
        Array.set res i (java_string_of_string (StdLabels.Array.get a i))
      done;
      res

    let bool_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_boolean a i)

    let byte_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_byte a i)

    let char_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> char_of_java_char (Array.get_char a i))

    let double_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_double a i)

    let float_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_float a i)

    let int_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_int a i)

    let long_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_long a i)

    let short_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> Array.get_short a i)

    let string_array_from_java a =
      let len = Array.get_length a in
      StdLabels.Array.init len ~f:(fun i -> string_of_java_string (Array.get a i))
  end


(* Exceptions *)

exception Java_exception of (string * string * java_object)

external print_stack_trace : out_channel -> java_object -> unit =
  "cadmium_print_stack_trace"

external throw_exception : java_object -> unit =
  "cadmium_throw_exception"

let _ =
  Callback.register_exception
    "Cadmium.Java_exception"
    (Java_exception ("", "", (Obj.magic 0 :> java_object)))


(* Access to Graphics elements *)

external get_frame : unit -> java_object =
  "cadmium_get_frame"

external get_canvas : unit -> java_object =
  "cadmium_get_canvas"

external get_buffer : unit -> java_object =
  "cadmium_get_buffer"


(* Callbacks *)

external create_interface : (string * (java_object java_type list) * java_object java_type) list -> java_class =
  "cadmium_create_interface"

external create_named_interface : (string * (java_object java_type list) * java_object java_type) list -> string -> java_class =
  "cadmium_create_named_interface"

external register : java_class list -> (java_method * (java_object java_value list -> java_object java_value)) list -> java_object =
  "cadmium_register"


(* Synchronization *)

external synchronized : java_object -> (unit -> unit) -> unit =
  "cadmium_synchronized"


(* Applet *)

type applet_parameter = string * string * string

type applet_handler = {
    info : string;
    parameters : applet_parameter array;
    init : java_object -> unit;
    start : java_object -> unit;
    run : java_object -> unit;
    stop : java_object -> unit;
    destroy : java_object -> unit;
  }

let register_applet h =
  Callback.register "Cadmium.applet.information" h.info;
  Callback.register "Cadmium.applet.parameters" h.parameters;
  Callback.register "Cadmium.applet.init" h.init;
  Callback.register "Cadmium.applet.start" h.start;
  Callback.register "Cadmium.applet.run" h.run;
  Callback.register "Cadmium.applet.stop" h.stop;
  Callback.register "Cadmium.applet.destroy" h.destroy

external get_applet_parameter : java_object -> string -> string =
  "cadmium_applet_get_parameter"


(* Runtime *)

type interpreter_parameters = {
    file : string;
    arguments : string array;
    init_stack_size : int;
    max_stack_size : int;
    backtrace : bool;
    exit_stops_jvm : bool;
    providers : string array;
    compile_dispatchers : bool;
    use_awt : bool;
    bare_canvas : bool;
    use_javax_sound : bool;
    use_jdbm : bool;
    os : string;
    unix_emulation : bool
  }

type interpreter = java_object

external is_native : unit -> bool =
  "cadmium_is_native"

external get_coderunner : unit -> java_object =
  "cadmium_get_coderunner"

external get_context : unit -> java_object =
  "cadmium_get_context"

external get_parameters : unit -> interpreter_parameters =
  "cadmium_get_parameters"

external make_interpreter : interpreter_parameters -> string -> string -> in_channel -> out_channel -> out_channel -> interpreter =
  "cadmium_make_interpreter" "cadmium_make_interpreter_native"

external interpreter_execute : interpreter -> 'a =
  "cadmium_interpreter_execute"

external interpreter_execute_callback : interpreter -> string -> 'a list -> 'b =
  "cadmium_interpreter_execute_callback"

external interpreter_get_context : interpreter -> java_object =
  "cadmium_interpreter_get_context"

external interpreter_get_parameters : interpreter -> interpreter_parameters =
  "cadmium_interpreter_get_parameters"

let java_object_of_interpreter x = x


(* Scripts *)

external get_binding : string -> 'a =
  "cadmium_get_binding"
