(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Types *)

type jdbc_type =
  | Array
  | Big_int
  | Binary
  | Bit
  | Blob
  | Boolean
  | Char
  | Clob
  | Datalink
  | Date
  | Decimal
  | Distinct
  | Double
  | Float
  | Integer
  | Java_object
  | Long_nvarchar
  | Long_varbinary
  | Long_varchar
  | Nchar
  | Nclob
  | Numeric
  | Nvarchar
  | Other
  | Real
  | Ref
  | Rowid
  | Small_int
  | SQL_XML
  | Struct
  | Time
  | Timestamp
  | Tiny_int
  | Varbinary
  | Varchar

type client_info_status =
  | Reason_unknown
  | Reason_unknown_property
  | Reason_value_invalid
  | Reason_value_truncated

type rowid_lifetime =
  | Rowid_unsupported
  | Rowid_valid_forever
  | Rowid_valid_other
  | Rowid_valid_session
  | Rowid_valid_transaction

type transaction_isolation =
  | Transaction_none
  | Transaction_read_committed
  | Transaction_read_uncommitted
  | Transaction_repeatable_read
  | Transaction_serializable

type result_set_concurrency =
  | Concurrency_read_only
  | Concurrency_updatable

type result_set_fetch_direction =
  | Fetch_forward
  | Fetch_reverse
  | Fetch_unknown

type result_set_holdability =
  | Close_cursors_at_commit
  | Hold_cursors_over_commit

type result_set_type =
  | Type_forward_only
  | Type_scroll_insensitive
  | Type_scroll_sensitive

type result_scope =
  | Scope_temporary
  | Scope_transaction
  | Scope_session

type statement_action =
  | Close_all_results
  | Close_current_result
  | Keep_current_result

type statement_autogen_keys =
  | No_generated_keys
  | Return_generated_keys

type statement_result =
  | Execute_failed
  | Success_no_info
  | Success of int32

type parameter_mode =
  | Mode_in
  | Mode_out
  | Mode_inout
  | Mode_unknown

type nullable =
  | No_nulls
  | Nullable
  | Nullable_unknown

type sql_state_type =
  | State_xopen
  | State_sql
  | State_sql99

module Savepoint =
  struct
    type t

    external get_savepoint_id : t -> int32 =
      "cadmiumjdbc_savepoint_get_savepoint_id"

    external get_savepoint_name : t -> string =
      "cadmiumjdbc_savepoint_get_savepoint_name"
  end

type sql_warning = { reason : string; state : string }


(* Data types *)

module Array =
  struct
    type t

    external free : t -> unit =
      "cadmiumjdbc_array_free"

    external get_array : t -> Cadmium.java_object =
      "cadmiumjdbc_array_get_array"

    external get_array_slice : t -> int64 -> int32 -> Cadmium.java_object =
      "cadmiumjdbc_array_get_array_slice"

    external get_base_type : t -> jdbc_type =
      "cadmiumjdbc_array_get_base_type"

    external get_base_type_name : t -> string =
      "cadmiumjdbc_array_get_base_type_name"
  end

module Blob =
  struct
    type t

    external free : t -> unit =
      "cadmiumjdbc_blob_free"

    external get_binary_stream : t -> in_channel =
      "cadmiumjdbc_blob_get_binary_stream"

    external get_binary_stream_pos : t -> int64 -> int64 -> in_channel =
      "cadmiumjdbc_blob_get_binary_stream_pos"

    external get_bytes : t -> int64 -> int32 -> string =
      "cadmiumjdbc_blob_get_bytes"

    external length : t -> int64 =
      "cadmiumjdbc_blob_length"

    external position_blob : t -> t -> int64 -> int64 =
      "cadmiumjdbc_blob_position_blob"

    external position_bytes : t -> string -> int64 -> int64 =
      "cadmiumjdbc_blob_position_bytes"

    external set_binary_stream : t -> int64 -> out_channel =
      "cadmiumjdbc_blob_set_binary_stream"

    external set_bytes : t -> int64 -> string -> int32 =
      "cadmiumjdbc_blob_set_bytes"

    external set_bytes_ofs : t -> int64 -> string -> int -> int -> int32 =
      "cadmiumjdbc_blob_set_bytes_ofs"

    external truncate : t -> int64 -> unit =
      "cadmiumjdbc_blob_truncate"
  end

module Clob =
  struct
    type t

    external free : t -> unit =
      "cadmiumjdbc_clob_free"

    external get_ascii_stream : t -> in_channel =
      "cadmiumjdbc_clob_get_ascii_stream"

    external get_sub_string : t -> int64 -> int32 -> string =
      "cadmiumjdbc_clob_get_sub_string"

    external length : t -> int64 =
      "cadmiumjdbc_clob_length"

    external position_clob : t -> t -> int64 -> int64 =
      "cadmiumjdbc_clob_position_clob"

    external position_string : t -> string -> int64 -> int64 =
      "cadmiumjdbc_clob_position_string"

    external set_ascii_stream : t -> int64 -> out_channel =
      "cadmiumjdbc_clob_set_ascii_stream"

    external set_string : t -> int64 -> string -> int32 =
      "cadmiumjdbc_clob_set_string"

    external set_string_ofs : t -> int64 -> string -> int -> int -> int32 =
      "cadmiumjdbc_clob_set_string_ofs"

    external truncate : t -> int64 -> unit =
      "cadmiumjdbc_clob_truncate"
  end

module Date =
  struct
    type t

    external make : unit -> t =
      "cadmiumjdbc_date_make"

    external make_of_int64 : int64 -> t =
      "cadmiumjdbc_date_make_of_int64"

    external make_of_string : string -> t =
      "cadmiumjdbc_date_make_of_string"

    external after : t -> t -> bool =
      "cadmiumjdbc_datebase_after"

    external before : t -> t -> bool =
      "cadmiumjdbc_datebase_before"

    external compare_to : t -> t -> int =
      "cadmiumjdbc_datebase_before"

    external equals : t -> t -> bool =
      "cadmiumjdbc_datebase_equals"

    external get_day_of_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_month"

    external get_day_of_week : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week"

    external get_day_of_week_in_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week_in_month"

    external get_day_of_year : t -> int =
      "cadmiumjdbc_datebase_get_day_of_year"

    external get_hour_of_day : t -> int =
      "cadmiumjdbc_datebase_get_hour_of_day"

    external get_millisecond : t -> int =
      "cadmiumjdbc_datebase_get_millisecond"

    external get_minute : t -> int =
      "cadmiumjdbc_datebase_get_minute"

    external get_month : t -> int =
      "cadmiumjdbc_datebase_get_month"

    external get_second : t -> int =
      "cadmiumjdbc_datebase_get_second"

    external get_time : t -> int64 =
      "cadmiumjdbc_date_get_time"

    external get_week_of_month : t -> int =
      "cadmiumjdbc_datebase_get_week_of_month"

    external get_week_of_year : t -> int =
      "cadmiumjdbc_datebase_get_week_of_year"

    external get_year : t -> int =
      "cadmiumjdbc_datebase_get_year"

    external set_day_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_month"

    external set_day_of_week : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week"

    external set_day_of_week_in_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week_in_month"

    external set_day_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_year"

    external set_hour_of_day : t -> int -> unit =
      "cadmiumjdbc_datebase_set_hour_of_day"

    external set_millisecond : t -> int -> unit =
      "cadmiumjdbc_datebase_set_millisecond"

    external set_minute : t -> int -> unit =
      "cadmiumjdbc_datebase_set_minute"

    external set_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_month"

    external set_second : t -> int -> unit =
      "cadmiumjdbc_datebase_set_second"

    external set_time : t -> int64 -> unit =
      "cadmiumjdbc_date_set_time"

    external set_week_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_month"

    external set_week_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_year"

    external set_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_year"

    external to_string : t -> string =
      "cadmiumjdbc_date_to_string"
  end

module NClob =
  struct
    type t

    external free : t -> unit =
      "cadmiumjdbc_clob_free"

    external get_ascii_stream : t -> in_channel =
      "cadmiumjdbc_clob_get_ascii_stream"

    external get_sub_string : t -> int64 -> int32 -> string =
      "cadmiumjdbc_clob_get_sub_string"

    external length : t -> int64 =
      "cadmiumjdbc_clob_length"

    external position_clob : t -> t -> int64 -> int64 =
      "cadmiumjdbc_clob_position_clob"

    external position_string : t -> string -> int64 -> int64 =
      "cadmiumjdbc_clob_position_string"

    external set_ascii_stream : t -> int64 -> out_channel =
      "cadmiumjdbc_clob_set_ascii_stream"

    external set_string : t -> int64 -> string -> int32 =
      "cadmiumjdbc_clob_set_string"

    external set_string_ofs : t -> int64 -> string -> int -> int -> int32 =
      "cadmiumjdbc_clob_set_string_ofs"

    external truncate : t -> int64 -> unit =
      "cadmiumjdbc_clob_truncate"
  end

module Ref =
  struct
    type t

    external get_base_type_name : t -> string =
      "cadmiumjdbc_ref_get_base_type_name"

    external get_object : t -> Cadmium.java_object =
      "cadmiumjdbc_ref_get_object"

    external set_object : t -> Cadmium.java_object -> unit =
      "cadmiumjdbc_ref_set_object"
  end

module RowId =
  struct
    type t

    external equals : t -> t -> bool =
      "cadmiumjdbc_rowid_equals"

    external get_bytes : t -> string =
      "cadmiumjdbc_rowid_get_bytes"

    external hash_code : t -> int32 =
      "cadmiumjdbc_rowid_hash_code"

    external to_string : t -> string =
      "cadmiumjdbc_rowid_to_string"
  end

module SQLXML =
  struct
    type t

    external free : t -> unit =
      "cadmiumjdbc_sqlxml_free"

    external get_binary_stream : t -> in_channel =
      "cadmiumjdbc_sqlxml_get_binary_stream"

    external get_string : t -> string =
      "cadmiumjdbc_sqlxml_get_string"

    external set_binary_stream : t -> out_channel =
      "cadmiumjdbc_sqlxml_set_binary_stream"

    external set_string : t -> string -> unit =
      "cadmiumjdbc_sqlxml_set_string"
  end

module Struct =
  struct
    type t

    external get_attributes : t -> Cadmium.java_object list =
      "cadmiumjdbc_struct_get_attributes"

    external get_sql_type_name : t -> string =
      "cadmiumjdbc_struct_get_sql_type_name"
  end

module Time =
  struct
    type t

    external make : unit -> t =
      "cadmiumjdbc_time_make"

    external make_of_int64 : int64 -> t =
      "cadmiumjdbc_time_make_of_int64"

    external make_of_string : string -> t =
      "cadmiumjdbc_time_make_of_string"

    external after : t -> t -> bool =
      "cadmiumjdbc_datebase_after"

    external before : t -> t -> bool =
      "cadmiumjdbc_datebase_before"

    external compare_to : t -> t -> int =
      "cadmiumjdbc_datebase_before"

    external equals : t -> t -> bool =
      "cadmiumjdbc_datebase_equals"

    external get_day_of_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_month"

    external get_day_of_week : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week"

    external get_day_of_week_in_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week_in_month"

    external get_day_of_year : t -> int =
      "cadmiumjdbc_datebase_get_day_of_year"

    external get_hour_of_day : t -> int =
      "cadmiumjdbc_datebase_get_hour_of_day"

    external get_millisecond : t -> int =
      "cadmiumjdbc_datebase_get_millisecond"

    external get_minute : t -> int =
      "cadmiumjdbc_datebase_get_minute"

    external get_month : t -> int =
      "cadmiumjdbc_datebase_get_month"

    external get_second : t -> int =
      "cadmiumjdbc_datebase_get_second"

    external get_time : t -> int64 =
      "cadmiumjdbc_time_get_time"

    external get_week_of_month : t -> int =
      "cadmiumjdbc_datebase_get_week_of_month"

    external get_week_of_year : t -> int =
      "cadmiumjdbc_datebase_get_week_of_year"

    external get_year : t -> int =
      "cadmiumjdbc_datebase_get_year"

    external set_day_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_month"

    external set_day_of_week : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week"

    external set_day_of_week_in_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week_in_month"

    external set_day_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_year"

    external set_hour_of_day : t -> int -> unit =
      "cadmiumjdbc_datebase_set_hour_of_day"

    external set_millisecond : t -> int -> unit =
      "cadmiumjdbc_datebase_set_millisecond"

    external set_minute : t -> int -> unit =
      "cadmiumjdbc_datebase_set_minute"

    external set_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_month"

    external set_second : t -> int -> unit =
      "cadmiumjdbc_datebase_set_second"

    external set_time : t -> int64 -> unit =
      "cadmiumjdbc_time_set_time"

    external set_week_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_month"

    external set_week_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_year"

    external set_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_year"

    external to_string : t -> string =
      "cadmiumjdbc_time_to_string"
  end

module Timestamp =
  struct
    type t

    external make : unit -> t =
      "cadmiumjdbc_timestamp_make"

    external make_of_int64 : int64 -> t =
      "cadmiumjdbc_timestamp_make_of_int64"

    external make_of_string : string -> t =
      "cadmiumjdbc_timestamp_make_of_string"

    external after : t -> t -> bool =
      "cadmiumjdbc_timestamp_after"

    external before : t -> t -> bool =
      "cadmiumjdbc_timestamp_before"

    external compare_to : t -> t -> int =
      "cadmiumjdbc_timestamp_compare_to"

    external equals : t -> t -> bool =
      "cadmiumjdbc_datebase_equals"

    external get_day_of_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_month"

    external get_day_of_week : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week"

    external get_day_of_week_in_month : t -> int =
      "cadmiumjdbc_datebase_get_day_of_week_in_month"

    external get_day_of_year : t -> int =
      "cadmiumjdbc_datebase_get_day_of_year"

    external get_hour_of_day : t -> int =
      "cadmiumjdbc_datebase_get_hour_of_day"

    external get_millisecond : t -> int =
      "cadmiumjdbc_datebase_get_millisecond"

    external get_minute : t -> int =
      "cadmiumjdbc_datebase_get_minute"

    external get_month : t -> int =
      "cadmiumjdbc_datebase_get_month"

    external get_nanos : t -> int32 =
      "cadmiumjdbc_timestamp_get_nanos"

    external get_second : t -> int =
      "cadmiumjdbc_datebase_get_second"

    external get_time : t -> int64 =
      "cadmiumjdbc_timestamp_get_time"

    external get_week_of_month : t -> int =
      "cadmiumjdbc_datebase_get_week_of_month"

    external get_week_of_year : t -> int =
      "cadmiumjdbc_datebase_get_week_of_year"

    external get_year : t -> int =
      "cadmiumjdbc_datebase_get_year"

    external set_day_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_month"

    external set_day_of_week : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week"

    external set_day_of_week_in_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_week_in_month"

    external set_day_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_day_of_year"

    external set_hour_of_day : t -> int -> unit =
      "cadmiumjdbc_datebase_set_hour_of_day"

    external set_millisecond : t -> int -> unit =
      "cadmiumjdbc_datebase_set_millisecond"

    external set_minute : t -> int -> unit =
      "cadmiumjdbc_datebase_set_minute"

    external set_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_month"

    external set_nanos : t -> int32 -> unit =
      "cadmiumjdbc_timestamp_set_nanos"

    external set_second : t -> int -> unit =
      "cadmiumjdbc_datebase_set_second"

    external set_time : t -> int64 -> unit =
      "cadmiumjdbc_timestamp_set_time"

    external set_week_of_month : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_month"

    external set_week_of_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_week_of_year"

    external set_year : t -> int -> unit =
      "cadmiumjdbc_datebase_set_year"

    external to_string : t -> string =
      "cadmiumjdbc_timestamp_to_string"
  end


(* Result sets *)

module ResultSetMetaData =
  struct
    type t

    external get_catalog_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_catalog_name"

    external get_column_class_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_column_class_name"

    external get_column_count : t -> int =
      "cadmiumjdbc_resultsetmetadata_get_column_count"

    external get_column_display_size : t -> int -> int =
      "cadmiumjdbc_resultsetmetadata_get_column_display_size"

    external get_column_label : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_column_label"

    external get_column_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_column_name"

    external get_column_type : t -> int -> jdbc_type =
      "cadmiumjdbc_resultsetmetadata_get_column_type"

    external get_column_type_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_column_type_name"

    external get_precision : t -> int -> int =
      "cadmiumjdbc_resultsetmetadata_get_precision"

    external get_scale : t -> int -> int =
      "cadmiumjdbc_resultsetmetadata_get_scale"

    external get_schema_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_schema_name"

    external get_table_name : t -> int -> string =
      "cadmiumjdbc_resultsetmetadata_get_table_name"

    external is_auto_increment : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_auto_increment"

    external is_case_sensitive : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_case_sensitive"

    external is_currency : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_currency"

    external is_definitely_writable : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_definitely_writable"

    external is_nullable : t -> int -> nullable =
      "cadmiumjdbc_resultsetmetadata_is_nullable"

    external is_read_only : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_read_only"

    external is_searchable : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_searchable"

    external is_signed : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_signed"

    external is_writable : t -> int -> bool =
      "cadmiumjdbc_resultsetmetadata_is_writable"
  end

module ResultSet =
  struct
    type t

    external absolute : t -> int32 -> bool =
      "cadmiumjdbc_resultset_absolute"

    external after_last : t -> unit =
      "cadmiumjdbc_resultset_after_last"

    external before_first : t -> unit =
      "cadmiumjdbc_resultset_before_first"

    external cancel_row_updates : t -> unit =
      "cadmiumjdbc_resultset_cancel_row_updates"

    external clear_warnings : t -> unit =
      "cadmiumjdbc_resultset_clear_warnings"

    external close : t -> unit =
      "cadmiumjdbc_resultset_close"

    external delete_row : t -> unit =
      "cadmiumjdbc_resultset_delete_row"

    external find_column : t -> string -> int =
      "cadmiumjdbc_resultset_find_column"

    external first : t -> bool =
      "cadmiumjdbc_resultset_first"

    external get_array_idx : t -> int -> Array.t option =
      "cadmiumjdbc_resultset_get_array_idx"

    external get_array_name : t -> string -> Array.t option =
      "cadmiumjdbc_resultset_get_array_name"

    external get_ascii_stream_idx : t -> int -> in_channel option =
      "cadmiumjdbc_resultset_get_ascii_stream_idx"

    external get_ascii_stream_name : t -> string -> in_channel option =
      "cadmiumjdbc_resultset_get_ascii_stream_name"

    external get_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option =
      "cadmiumjdbc_resultset_get_big_decimal_idx"

    external get_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option =
      "cadmiumjdbc_resultset_get_big_decimal_name"

    external get_binary_stream_idx : t -> int -> in_channel option =
      "cadmiumjdbc_resultset_get_binary_stream_idx"

    external get_binary_stream_name : t -> string -> in_channel option =
      "cadmiumjdbc_resultset_get_binary_stream_name"

    external get_blob_idx : t -> int -> Blob.t option =
      "cadmiumjdbc_resultset_get_blob_idx"

    external get_blob_name : t -> string -> Blob.t option =
      "cadmiumjdbc_resultset_get_blob_name"

    external get_boolean_idx : t -> int -> bool option =
      "cadmiumjdbc_resultset_get_boolean_idx"

    external get_boolean_name : t -> string -> bool option =
      "cadmiumjdbc_resultset_get_boolean_name"

    external get_byte_idx : t -> int -> int option =
      "cadmiumjdbc_resultset_get_byte_idx"

    external get_byte_name : t -> string -> int option =
      "cadmiumjdbc_resultset_get_byte_name"

    external get_bytes_idx : t -> int -> string option =
      "cadmiumjdbc_resultset_get_bytes_idx"

    external get_bytes_name : t -> string -> string option =
      "cadmiumjdbc_resultset_get_bytes_name"

    external get_clob_idx : t -> int -> Clob.t option =
      "cadmiumjdbc_resultset_get_clob_idx"

    external get_clob_name : t -> string -> Clob.t option =
      "cadmiumjdbc_resultset_get_clob_name"

    external get_concurrency : t -> result_set_concurrency =
      "cadmiumjdbc_resultset_get_concurrency"

    external get_cursor_name : t -> string =
      "cadmiumjdbc_resultset_get_cursor_name"

    external get_date_idx : t -> int -> Date.t option =
      "cadmiumjdbc_resultset_get_date_idx"

    external get_date_name : t -> string -> Date.t option =
      "cadmiumjdbc_resultset_get_date_name"

    external get_double_idx : t -> int -> float option =
      "cadmiumjdbc_resultset_get_double_idx"

    external get_double_name : t -> string -> float option =
      "cadmiumjdbc_resultset_get_double_name"

    external get_fetch_direction : t -> result_set_fetch_direction =
      "cadmiumjdbc_resultset_get_fetch_direction"

    external get_fetch_size : t -> int32 =
      "cadmiumjdbc_resultset_get_fetch_size"

    external get_float_idx : t -> int -> float option =
      "cadmiumjdbc_resultset_get_float_idx"

    external get_float_name : t -> string -> float option =
      "cadmiumjdbc_resultset_get_float_name"

    external get_holdability : t -> result_set_holdability =
      "cadmiumjdbc_resultset_get_holdability"

    external get_int_idx : t -> int -> int32 option =
      "cadmiumjdbc_resultset_get_int_idx"

    external get_int_name : t -> string -> int32 option =
      "cadmiumjdbc_resultset_get_int_name"

    external get_long_idx : t -> int -> int64 option =
      "cadmiumjdbc_resultset_get_long_idx"

    external get_long_name : t -> string -> int64 option =
      "cadmiumjdbc_resultset_get_long_name"

    external get_meta_data : t -> ResultSetMetaData.t =
      "cadmiumjdbc_resultset_get_meta_data"

    external get_nclob_idx : t -> int -> NClob.t option =
      "cadmiumjdbc_resultset_get_nclob_idx"

    external get_nclob_name : t -> string -> NClob.t option =
      "cadmiumjdbc_resultset_get_nclob_name"

    external get_nstring_idx : t -> int -> string option =
      "cadmiumjdbc_resultset_get_nstring_idx"

    external get_nstring_name : t -> string -> string option =
      "cadmiumjdbc_resultset_get_nstring_name"

    external get_object_idx : t -> int -> Cadmium.java_object option =
      "cadmiumjdbc_resultset_get_object_idx"

    external get_object_name : t -> string -> Cadmium.java_object option =
      "cadmiumjdbc_resultset_get_object_name"

    external get_ref_idx : t -> int -> Ref.t option =
      "cadmiumjdbc_resultset_get_ref_idx"

    external get_ref_name : t -> string -> Ref.t option =
      "cadmiumjdbc_resultset_get_ref_name"

    external get_row : t -> int32 =
      "cadmiumjdbc_resultset_get_row"

    external get_row_id_idx : t -> int -> RowId.t option =
      "cadmiumjdbc_resultset_get_row_id_idx"

    external get_row_id_name : t -> string -> RowId.t option =
      "cadmiumjdbc_resultset_get_row_id_name"

    external get_short_idx : t -> int -> int option =
      "cadmiumjdbc_resultset_get_short_idx"

    external get_short_name : t -> string -> int option =
      "cadmiumjdbc_resultset_get_short_name"

    external get_sqlxml_idx : t -> int -> SQLXML.t option =
      "cadmiumjdbc_resultset_get_sqlxml_idx"

    external get_sqlxml_name : t -> string -> SQLXML.t option =
      "cadmiumjdbc_resultset_get_sqlxml_name"

    external get_string_idx : t -> int -> string option =
      "cadmiumjdbc_resultset_get_string_idx"

    external get_string_name : t -> string -> string option =
      "cadmiumjdbc_resultset_get_string_name"

    external get_time_idx : t -> int -> Time.t option =
      "cadmiumjdbc_resultset_get_time_idx"

    external get_time_name : t -> string -> Time.t option =
      "cadmiumjdbc_resultset_get_time_name"

    external get_timestamp_idx : t -> int -> Timestamp.t option =
      "cadmiumjdbc_resultset_get_timestamp_idx"

    external get_timestamp_name : t -> string -> Timestamp.t option =
      "cadmiumjdbc_resultset_get_timestamp_name"

    external get_type : t -> result_set_type =
      "cadmiumjdbc_resultset_get_type"

    external get_url_idx : t -> int -> string option =
      "cadmiumjdbc_resultset_get_url_idx"

    external get_url_name : t -> string -> string option =
      "cadmiumjdbc_resultset_get_url_name"

    external get_warnings : t -> sql_warning list =
      "cadmiumjdbc_resultset_get_warnings"

    external insert_row : t -> unit =
      "cadmiumjdbc_resultset_insert_row"

    external is_after_last : t -> bool =
      "cadmiumjdbc_resultset_is_after_last"

    external is_before_first : t -> bool =
      "cadmiumjdbc_resultset_is_before_first"

    external is_closed : t -> bool =
      "cadmiumjdbc_resultset_is_closed"

    external is_first : t -> bool =
      "cadmiumjdbc_resultset_is_first"

    external is_last : t -> bool =
      "cadmiumjdbc_resultset_is_last"

    external last : t -> bool =
      "cadmiumjdbc_resultset_last"

    external move_to_current_row : t -> unit =
      "cadmiumjdbc_resultset_move_to_current_row"

    external move_to_insert_row : t -> unit =
      "cadmiumjdbc_resultset_move_to_insert_row"

    external next : t -> bool =
      "cadmiumjdbc_resultset_next"

    external previous : t -> bool =
      "cadmiumjdbc_resultset_previous"

    external refresh_row : t -> unit =
      "cadmiumjdbc_resultset_refresh_row"

    external relative : t -> int32 -> bool =
      "cadmiumjdbc_resultset_relative"

    external row_deleted : t -> bool =
      "cadmiumjdbc_resultset_row_deleted"

    external row_inserted : t -> bool =
      "cadmiumjdbc_resultset_row_inserted"

    external row_updated : t -> bool =
      "cadmiumjdbc_resultset_row_updated"

    external set_fetch_direction : t -> result_set_fetch_direction -> unit =
      "cadmiumjdbc_resultset_set_fetch_direction"

    external set_fetch_size : t -> int32 -> unit =
      "cadmiumjdbc_resultset_set_fetch_size"

    external update_array_idx : t -> int -> Array.t option -> unit =
      "cadmiumjdbc_resultset_update_array_idx"

    external update_array_name : t -> string -> Array.t option -> unit =
      "cadmiumjdbc_resultset_update_array_name"

    external update_ascii_stream_idx : t -> int -> in_channel option -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_idx"

    external update_ascii_stream_name : t -> string -> in_channel option -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_name"

    external update_ascii_stream_length_idx : t -> int -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_length_idx"

    external update_ascii_stream_length_name : t -> string -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_length_name"

    external update_ascii_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_length64_idx"

    external update_ascii_stream_length64_name : t -> string -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_resultset_update_ascii_stream_length64_name"

    external update_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option -> unit =
      "cadmiumjdbc_resultset_update_big_decimal_idx"

    external update_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option -> unit =
      "cadmiumjdbc_resultset_update_big_decimal_name"

    external update_binary_stream_idx : t -> int -> in_channel option -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_idx"

    external update_binary_stream_name : t -> string -> in_channel option -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_name"

    external update_binary_stream_length_idx : t -> int -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_length_idx"

    external update_binary_stream_length_name : t -> string -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_length_name"

    external update_binary_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_length64_idx"

    external update_binary_stream_length64_name : t -> string -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_resultset_update_binary_stream_length64_name"

    external update_blob_idx : t -> int -> Blob.t option -> unit =
      "cadmiumjdbc_resultset_update_blob_idx"

    external update_blob_name : t -> string -> Blob.t option -> unit =
      "cadmiumjdbc_resultset_update_blob_name"

    external update_boolean_idx : t -> int -> bool option -> unit =
      "cadmiumjdbc_resultset_update_boolean_idx"

    external update_boolean_name : t -> string -> bool option -> unit =
      "cadmiumjdbc_resultset_update_boolean_name"

    external update_byte_idx : t -> int -> int option -> unit =
      "cadmiumjdbc_resultset_update_byte_idx"

    external update_byte_name : t -> string -> int option -> unit =
      "cadmiumjdbc_resultset_update_byte_name"

    external update_bytes_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_resultset_update_bytes_idx"

    external update_bytes_name : t -> string -> string option -> unit =
      "cadmiumjdbc_resultset_update_bytes_name"

    external update_clob_idx : t -> int -> Clob.t option -> unit =
      "cadmiumjdbc_resultset_update_clob_idx"

    external update_clob_name : t -> string -> Clob.t option -> unit =
      "cadmiumjdbc_resultset_update_clob_name"

    external update_date_idx : t -> int -> Date.t option -> unit =
      "cadmiumjdbc_resultset_update_date_idx"

    external update_date_name : t -> string -> Date.t option -> unit =
      "cadmiumjdbc_resultset_update_date_name"

    external update_double_idx : t -> int -> float option -> unit =
      "cadmiumjdbc_resultset_update_double_idx"

    external update_double_name : t -> string -> float option -> unit =
      "cadmiumjdbc_resultset_update_double_name"

    external update_float_idx : t -> int -> float option -> unit =
      "cadmiumjdbc_resultset_update_float_idx"

    external update_float_name : t -> string -> float option -> unit =
      "cadmiumjdbc_resultset_update_float_name"

    external update_int_idx : t -> int -> int32 option -> unit =
      "cadmiumjdbc_resultset_update_int_idx"

    external update_int_name : t -> string -> int32 option -> unit =
      "cadmiumjdbc_resultset_update_int_name"

    external update_long_idx : t -> int -> int64 option -> unit =
      "cadmiumjdbc_resultset_update_long_idx"

    external update_long_name : t -> string -> int64 option -> unit =
      "cadmiumjdbc_resultset_update_long_name"

    external update_nclob_idx : t -> int -> NClob.t option -> unit =
      "cadmiumjdbc_resultset_update_nclob_idx"

    external update_nclob_name : t -> string -> NClob.t option -> unit =
      "cadmiumjdbc_resultset_update_nclob_name"

    external update_nstring_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_resultset_update_nstring_idx"

    external update_nstring_name : t -> string -> string option -> unit =
      "cadmiumjdbc_resultset_update_nstring_name"

    external update_object_idx : t -> int -> Cadmium.java_object option -> unit =
      "cadmiumjdbc_resultset_update_object_idx"

    external update_object_name : t -> string -> Cadmium.java_object option -> unit =
      "cadmiumjdbc_resultset_update_object_name"

    external update_object_scale_idx : t -> int -> Cadmium.java_object option -> int -> unit =
      "cadmiumjdbc_resultset_update_object_scale_idx"

    external update_object_scale_name : t -> string -> Cadmium.java_object option -> int -> unit =
      "cadmiumjdbc_resultset_update_object_scale_name"

    external update_ref_idx : t -> int -> Ref.t option -> unit =
      "cadmiumjdbc_resultset_update_ref_idx"

    external update_ref_name : t -> string -> Ref.t option -> unit =
      "cadmiumjdbc_resultset_update_ref_name"

    external update_row : t -> unit =
      "cadmiumjdbc_resultset_update_row"

    external update_row_id_idx : t -> int -> RowId.t option -> unit =
      "cadmiumjdbc_resultset_update_row_id_idx"

    external update_row_id_name : t -> string -> RowId.t option -> unit =
      "cadmiumjdbc_resultset_update_row_id_name"

    external update_short_idx : t -> int -> int option -> unit =
      "cadmiumjdbc_resultset_update_short_idx"

    external update_short_name : t -> string -> int option -> unit =
      "cadmiumjdbc_resultset_update_short_name"

    external update_sqlxml_idx : t -> int -> SQLXML.t option -> unit =
      "cadmiumjdbc_resultset_update_sqlxml_idx"

    external update_sqlxml_name : t -> string -> SQLXML.t option -> unit =
      "cadmiumjdbc_resultset_update_sqlxml_name"

    external update_string_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_resultset_update_string_idx"

    external update_string_name : t -> string -> string option -> unit =
      "cadmiumjdbc_resultset_update_string_name"

    external update_time_idx : t -> int -> Time.t option -> unit =
      "cadmiumjdbc_resultset_update_time_idx"

    external update_time_name : t -> string -> Time.t option -> unit =
      "cadmiumjdbc_resultset_update_time_name"

    external update_timestamp_idx : t -> int -> Timestamp.t option -> unit =
      "cadmiumjdbc_resultset_update_timestamp_idx"

    external update_timestamp_name : t -> string -> Timestamp.t option -> unit =
      "cadmiumjdbc_resultset_update_timestamp_name"
  end


(* Statements *)

module ParameterMetaData =
  struct
    type t

    external get_parameter_class_name : t -> int -> string =
      "cadmiumjdbc_parametermetadata_get_parameter_class_name"

    external get_parameter_count : t -> int =
      "cadmiumjdbc_parametermetadata_get_parameter_count"

    external get_parameter_mode : t -> int -> parameter_mode =
      "cadmiumjdbc_parametermetadata_get_parameter_mode"

    external get_parameter_type : t -> int -> jdbc_type =
      "cadmiumjdbc_parametermetadata_get_parameter_type"

    external get_parameter_type_name : t -> int -> string =
      "cadmiumjdbc_parametermetadata_get_parameter_type_name"

    external get_precision : t -> int -> int =
      "cadmiumjdbc_parametermetadata_get_precision"

    external get_scale : t -> int -> int =
      "cadmiumjdbc_parametermetadata_get_scale"

    external is_nullable : t -> int -> nullable =
      "cadmiumjdbc_parametermetadata_is_nullable"

    external is_signed : t -> int -> bool =
      "cadmiumjdbc_parametermetadata_is_signed"
  end

module Statement =
  struct
    type t

    external add_batch : t -> string -> unit =
      "cadmiumjdbc_statement_add_batch"

    external cancel : t -> unit =
      "cadmiumjdbc_statement_cancel"

    external clear_batch : t -> unit =
      "cadmiumjdbc_statement_clear_batch"

    external clear_warnings : t -> unit =
      "cadmiumjdbc_statement_clear_warnings"

    external close : t -> unit =
      "cadmiumjdbc_statement_close"

    external execute : t -> string -> bool =
      "cadmiumjdbc_statement_execute"

    external execute_autogen : t -> string -> statement_autogen_keys -> bool =
      "cadmiumjdbc_statement_execute_autogen"

    external execute_idx : t -> string -> int list -> bool =
      "cadmiumjdbc_statement_execute_idx"

    external execute_name : t -> string -> string list -> bool =
      "cadmiumjdbc_statement_execute_name"

    external execute_batch : t -> statement_result list =
      "cadmiumjdbc_statement_execute_batch"

    external execute_query : t -> string -> ResultSet.t =
      "cadmiumjdbc_statement_execute_query"

    external execute_update : t -> string -> int32 =
      "cadmiumjdbc_statement_execute_update"

    external execute_update_autogen : t -> string -> statement_autogen_keys -> int32 =
      "cadmiumjdbc_statement_execute_update_autogen"

    external execute_update_idx : t -> string -> int list -> int32 =
      "cadmiumjdbc_statement_execute_update_idx"

    external execute_update_name : t -> string -> string list -> int32 =
      "cadmiumjdbc_statement_execute_update_name"

    external get_fetch_direction : t -> result_set_fetch_direction =
      "cadmiumjdbc_statement_get_fetch_direction"

    external get_fetch_size : t -> int32 =
      "cadmiumjdbc_statement_get_fetch_size"

    external get_generated_keys : t -> ResultSet.t =
      "cadmiumjdbc_statement_get_generated_keys"

    external get_max_field_size : t -> int32 =
      "cadmiumjdbc_statement_get_max_field_size"

    external get_max_rows : t -> int32 =
      "cadmiumjdbc_statement_get_max_rows"

    external get_more_results : t -> bool =
      "cadmiumjdbc_statement_get_more_results"

    external get_more_results_current : t -> statement_action -> bool =
      "cadmiumjdbc_statement_get_more_results_current"

    external get_query_timeout : t -> int32 =
      "cadmiumjdbc_statement_get_query_timeout"

    external get_result_set : t -> ResultSet.t =
      "cadmiumjdbc_statement_get_result_set"

    external get_result_set_concurrency : t -> result_set_concurrency =
      "cadmiumjdbc_statement_get_result_set_concurrency"

    external get_result_set_holdability : t -> result_set_holdability =
      "cadmiumjdbc_statement_get_result_set_holdability"

    external get_result_set_type : t -> result_set_type =
      "cadmiumjdbc_statement_get_result_set_type"

    external get_update_count : t -> int32 =
      "cadmiumjdbc_statement_get_update_count"

    external get_warnings : t -> sql_warning list =
      "cadmiumjdbc_statement_get_warnings"

    external is_closed : t -> bool =
      "cadmiumjdbc_statement_is_closed"

    external is_poolable : t -> bool =
      "cadmiumjdbc_statement_is_poolable"

    external set_cursor_name : t -> string -> unit =
      "cadmiumjdbc_statement_set_cursor_name"

    external set_escape_processing : t -> bool -> unit =
      "cadmiumjdbc_statement_set_escape_processing"

    external set_fetch_direction : t -> result_set_fetch_direction -> unit =
      "cadmiumjdbc_statement_set_fetch_direction"

    external set_fetch_size : t -> int32 -> unit =
      "cadmiumjdbc_statement_set_fetch_size"

    external set_max_field_size : t -> int32 -> unit =
      "cadmiumjdbc_statement_set_max_field_size"

    external set_max_rows : t -> int32 -> unit =
      "cadmiumjdbc_statement_set_max_rows"

    external set_poolable : t -> bool -> unit =
      "cadmiumjdbc_statement_set_poolable"

    external set_query_timeout : t -> int32 -> unit =
      "cadmiumjdbc_statement_set_query_timeout"
  end

module PreparedStatement =
  struct
    type t

    external as_statement : t -> Statement.t =
      "%identity"

    external add_batch : t -> unit =
      "cadmiumjdbc_preparedstatement_add_batch"

    external clear_parameters : t -> unit =
      "cadmiumjdbc_preparedstatement_clear_parameters"

    external execute : t -> bool =
      "cadmiumjdbc_preparedstatement_execute"

    external execute_query : t -> ResultSet.t =
      "cadmiumjdbc_preparedstatement_execute_query"

    external execute_update : t -> int32 =
      "cadmiumjdbc_preparedstatement_execute_update"

    external get_meta_data : t -> ResultSetMetaData.t =
      "cadmiumjdbc_preparedstatement_get_meta_data"

    external get_parameter_meta_data : t -> ParameterMetaData.t =
      "cadmiumjdbc_preparedstatement_get_parameter_meta_data"

    external set_array_idx : t -> int -> Array.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_array_idx"

    external set_ascii_stream_idx : t -> int -> in_channel option -> unit =
      "cadmiumjdbc_preparedstatement_set_ascii_stream_idx"

    external set_ascii_stream_length_idx : t -> int -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_preparedstatement_set_ascii_stream_length_idx"

    external set_ascii_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_preparedstatement_set_ascii_stream_length64_idx"

    external set_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_big_decimal_idx"

    external set_binary_stream_idx : t -> int -> in_channel option -> unit =
      "cadmiumjdbc_preparedstatement_set_binary_stream_idx"

    external set_binary_stream_length_idx : t -> int -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_preparedstatement_set_binary_stream_length_idx"

    external set_binary_stream_length64_idx : t -> int -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_preparedstatement_set_binary_stream_length64_idx"

    external set_blob_idx : t -> int -> Blob.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_blob_idx"

    external set_boolean_idx : t -> int -> bool option -> unit =
      "cadmiumjdbc_preparedstatement_set_boolean_idx"

    external set_byte_idx : t -> int -> int option -> unit =
      "cadmiumjdbc_preparedstatement_set_byte_idx"

    external set_bytes_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_preparedstatement_set_bytes_idx"

    external set_clob_idx : t -> int -> Clob.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_clob_idx"

    external set_date_idx : t -> int -> Date.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_date_idx"

    external set_double_idx : t -> int -> float option -> unit =
      "cadmiumjdbc_preparedstatement_set_double_idx"

    external set_float_idx : t -> int -> float option -> unit =
      "cadmiumjdbc_preparedstatement_set_float_idx"

    external set_int_idx : t -> int -> int32 option -> unit =
      "cadmiumjdbc_preparedstatement_set_int_idx"

    external set_long_idx : t -> int -> int64 option -> unit =
      "cadmiumjdbc_preparedstatement_set_long_idx"

    external set_nclob_idx : t -> int -> NClob.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_nclob_idx"

    external set_nstring_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_preparedstatement_set_nstring_idx"

    external set_object_idx : t -> int -> Cadmium.java_object option -> unit =
      "cadmiumjdbc_preparedstatement_set_object_idx"

    external set_object_type_idx : t -> int -> Cadmium.java_object option -> jdbc_type -> unit =
      "cadmiumjdbc_preparedstatement_set_object_type_idx"

    external set_object_type_scale_idx : t -> int -> Cadmium.java_object option -> jdbc_type -> int -> unit =
      "cadmiumjdbc_preparedstatement_set_object_type_scale_idx"

    external set_ref_idx : t -> int -> Ref.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_ref_idx"

    external set_row_id_idx : t -> int -> RowId.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_row_id_idx"

    external set_short_idx : t -> int -> int option -> unit =
      "cadmiumjdbc_preparedstatement_set_short_idx"

    external set_sqlxml_idx : t -> int -> SQLXML.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_sqlxml_idx"

    external set_string_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_preparedstatement_set_string_idx"

    external set_time_idx : t -> int -> Time.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_time_idx"

    external set_timestamp_idx : t -> int -> Timestamp.t option -> unit =
      "cadmiumjdbc_preparedstatement_set_timestamp_idx"

    external set_url_idx : t -> int -> string option -> unit =
      "cadmiumjdbc_preparedstatement_set_url_idx"
  end

module CallableStatement =
  struct
    type t

    external as_statement : t -> Statement.t =
      "%identity"

    external as_prepared_statement : t -> PreparedStatement.t =
      "%identity"

    external get_array_idx : t -> int -> Array.t option =
      "cadmiumjdbc_callablestatement_get_array_idx"

    external get_array_name : t -> string -> Array.t option =
      "cadmiumjdbc_callablestatement_get_array_name"

    external get_big_decimal_idx : t -> int -> CadmiumMath.BigDecimal.t option =
      "cadmiumjdbc_callablestatement_get_big_decimal_idx"

    external get_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option =
      "cadmiumjdbc_callablestatement_get_big_decimal_name"

    external get_blob_idx : t -> int -> Blob.t option =
      "cadmiumjdbc_callablestatement_get_blob_idx"

    external get_blob_name : t -> string -> Blob.t option =
      "cadmiumjdbc_callablestatement_get_blob_name"

    external get_boolean_idx : t -> int -> bool option =
      "cadmiumjdbc_callablestatement_get_boolean_idx"

    external get_boolean_name : t -> string -> bool option =
      "cadmiumjdbc_callablestatement_get_boolean_name"

    external get_byte_idx : t -> int -> int option =
      "cadmiumjdbc_callablestatement_get_byte_idx"

    external get_byte_name : t -> string -> int option =
      "cadmiumjdbc_callablestatement_get_byte_name"

    external get_bytes_idx : t -> int -> string option =
      "cadmiumjdbc_callablestatement_get_bytes_idx"

    external get_bytes_name : t -> string -> string option =
      "cadmiumjdbc_callablestatement_get_bytes_name"

    external get_clob_idx : t -> int -> Clob.t option =
      "cadmiumjdbc_callablestatement_get_clob_idx"

    external get_clob_name : t -> string -> Clob.t option =
      "cadmiumjdbc_callablestatement_get_clob_name"

    external get_date_idx : t -> int -> Date.t option =
      "cadmiumjdbc_callablestatement_get_date_idx"

    external get_date_name : t -> string -> Date.t option =
      "cadmiumjdbc_callablestatement_get_date_name"

    external get_double_idx : t -> int -> float option =
      "cadmiumjdbc_callablestatement_get_double_idx"

    external get_double_name : t -> string -> float option =
      "cadmiumjdbc_callablestatement_get_double_name"

    external get_float_idx : t -> int -> float option =
      "cadmiumjdbc_callablestatement_get_float_idx"

    external get_float_name : t -> string -> float option =
      "cadmiumjdbc_callablestatement_get_float_name"

    external get_int_idx : t -> int -> int32 option =
      "cadmiumjdbc_callablestatement_get_int_idx"

    external get_int_name : t -> string -> int32 option =
      "cadmiumjdbc_callablestatement_get_int_name"

    external get_long_idx : t -> int -> int64 option =
      "cadmiumjdbc_callablestatement_get_long_idx"

    external get_long_name : t -> string -> int64 option =
      "cadmiumjdbc_callablestatement_get_long_name"

    external get_nclob_idx : t -> int -> NClob.t option =
      "cadmiumjdbc_callablestatement_get_nclob_idx"

    external get_nclob_name : t -> string -> NClob.t option =
      "cadmiumjdbc_callablestatement_get_nclob_name"

    external get_nstring_idx : t -> int -> string option =
      "cadmiumjdbc_callablestatement_get_nstring_idx"

    external get_nstring_name : t -> string -> string option =
      "cadmiumjdbc_callablestatement_get_nstring_name"

    external get_object_idx : t -> int -> Cadmium.java_object option =
      "cadmiumjdbc_callablestatement_get_object_idx"

    external get_object_name : t -> string -> Cadmium.java_object option =
      "cadmiumjdbc_callablestatement_get_object_name"

    external get_ref_idx : t -> int -> Ref.t option =
      "cadmiumjdbc_callablestatement_get_ref_idx"

    external get_ref_name : t -> string -> Ref.t option =
      "cadmiumjdbc_callablestatement_get_ref_name"

    external get_row_id_idx : t -> int -> RowId.t option =
      "cadmiumjdbc_callablestatement_get_row_id_idx"

    external get_row_id_name : t -> string -> RowId.t option =
      "cadmiumjdbc_callablestatement_get_row_id_name"

    external get_short_idx : t -> int -> int option =
      "cadmiumjdbc_callablestatement_get_short_idx"

    external get_short_name : t -> string -> int option =
      "cadmiumjdbc_callablestatement_get_short_name"

    external get_sqlxml_idx : t -> int -> SQLXML.t option =
      "cadmiumjdbc_callablestatement_get_sqlxml_idx"

    external get_sqlxml_name : t -> string -> SQLXML.t option =
      "cadmiumjdbc_callablestatement_get_sqlxml_name"

    external get_string_idx : t -> int -> string option =
      "cadmiumjdbc_callablestatement_get_string_idx"

    external get_string_name : t -> string -> string option =
      "cadmiumjdbc_callablestatement_get_string_name"

    external get_time_idx : t -> int -> Time.t option =
      "cadmiumjdbc_callablestatement_get_time_idx"

    external get_time_name : t -> string -> Time.t option =
      "cadmiumjdbc_callablestatement_get_time_name"

    external get_timestamp_idx : t -> int -> Timestamp.t option =
      "cadmiumjdbc_callablestatement_get_timestamp_idx"

    external get_timestamp_name : t -> string -> Timestamp.t option =
      "cadmiumjdbc_callablestatement_get_timestamp_name"

    external get_url_idx : t -> int -> string option =
      "cadmiumjdbc_callablestatement_get_url_idx"

    external get_url_name : t -> string -> string option =
      "cadmiumjdbc_callablestatement_get_url_name"

    external register_out_parameter_idx : t -> int -> jdbc_type -> unit =
      "cadmiumjdbc_callablestatement_register_out_parameter_idx"

    external register_out_parameter_scale_idx : t -> int -> jdbc_type -> int -> unit =
      "cadmiumjdbc_callablestatement_register_out_parameter_scale_idx"

    external register_out_parameter_name : t -> string -> jdbc_type -> unit =
      "cadmiumjdbc_callablestatement_register_out_parameter_name"

    external register_out_parameter_scale_name : t -> string -> jdbc_type -> int -> unit =
      "cadmiumjdbc_callablestatement_register_out_parameter_scale_name"

    external set_ascii_stream_name : t -> string -> in_channel option -> unit =
      "cadmiumjdbc_callablestatement_set_ascii_stream_name"

    external set_ascii_stream_length_name : t -> string -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_callablestatement_set_ascii_stream_length_name"

    external set_ascii_stream_length64_name : t -> string -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_callablestatement_set_ascii_stream_length64_name"

    external set_big_decimal_name : t -> string -> CadmiumMath.BigDecimal.t option -> unit =
      "cadmiumjdbc_callablestatement_set_big_decimal_name"

    external set_binary_stream_name : t -> string -> in_channel option -> unit =
      "cadmiumjdbc_callablestatement_set_binary_stream_name"

    external set_binary_stream_length_name : t -> string -> in_channel option -> int32 -> unit =
      "cadmiumjdbc_callablestatement_set_binary_stream_length_name"

    external set_binary_stream_length64_name : t -> string -> in_channel option -> int64 -> unit =
      "cadmiumjdbc_callablestatement_set_binary_stream_length64_name"

    external set_blob_name : t -> string -> Blob.t option -> unit =
      "cadmiumjdbc_callablestatement_set_blob_name"

    external set_boolean_name : t -> string -> bool option -> unit =
      "cadmiumjdbc_callablestatement_set_boolean_name"

    external set_byte_name : t -> string -> int option -> unit =
      "cadmiumjdbc_callablestatement_set_byte_name"

    external set_bytes_name : t -> string -> string option -> unit =
      "cadmiumjdbc_callablestatement_set_bytes_name"

    external set_clob_name : t -> string -> Clob.t option -> unit =
      "cadmiumjdbc_callablestatement_set_clob_name"

    external set_date_name : t -> string -> Date.t option -> unit =
      "cadmiumjdbc_callablestatement_set_date_name"

    external set_double_name : t -> string -> float option -> unit =
      "cadmiumjdbc_callablestatement_set_double_name"

    external set_float_name : t -> string -> float option -> unit =
      "cadmiumjdbc_callablestatement_set_float_name"

    external set_int_name : t -> string -> int32 option -> unit =
      "cadmiumjdbc_callablestatement_set_int_name"

    external set_long_name : t -> string -> int64 option -> unit =
      "cadmiumjdbc_callablestatement_set_long_name"

    external set_nclob_name : t -> string -> NClob.t option -> unit =
      "cadmiumjdbc_callablestatement_set_nclob_name"

    external set_nstring_name : t -> string -> string option -> unit =
      "cadmiumjdbc_callablestatement_set_nstring_name"

    external set_object_name : t -> string -> Cadmium.java_object option -> unit =
      "cadmiumjdbc_callablestatement_set_object_name"

    external set_object_type_name : t -> string -> Cadmium.java_object option -> jdbc_type -> unit =
      "cadmiumjdbc_callablestatement_set_object_type_name"

    external set_object_type_scale_name : t -> string -> Cadmium.java_object option -> jdbc_type -> int -> unit =
      "cadmiumjdbc_callablestatement_set_object_type_scale_name"

    external set_row_id_name : t -> string -> RowId.t option -> unit =
      "cadmiumjdbc_callablestatement_set_row_id_name"

    external set_short_name : t -> string -> int option -> unit =
      "cadmiumjdbc_callablestatement_set_short_name"

    external set_sqlxml_name : t -> string -> SQLXML.t option -> unit =
      "cadmiumjdbc_callablestatement_set_sqlxml_name"

    external set_string_name : t -> string -> string option -> unit =
      "cadmiumjdbc_callablestatement_set_string_name"

    external set_time_name : t -> string -> Time.t option -> unit =
      "cadmiumjdbc_callablestatement_set_time_name"

    external set_timestamp_name : t -> string -> Timestamp.t option -> unit =
      "cadmiumjdbc_callablestatement_set_timestamp_name"

    external set_url_name : t -> string -> string option -> unit =
      "cadmiumjdbc_callablestatement_set_url_name"
  end


(* Connections & Drivers *)

module DatabaseMetaData =
  struct
    type t

    external all_procedures_are_callable : t -> bool =
      "cadmiumjdbc_databasemetadata_all_procedures_are_callable"

    external all_tables_are_selectable : t -> bool =
      "cadmiumjdbc_databasemetadata_all_tables_are_selectable"

    external auto_commit_failure_closes_all_result_sets : t -> bool =
      "cadmiumjdbc_databasemetadata_auto_commit_failure_closes_all_result_sets"

    external data_definition_causes_transaction_commit : t -> bool =
      "cadmiumjdbc_databasemetadata_data_definition_causes_transaction_commit"

    external data_definition_ignored_in_transactions : t -> bool =
      "cadmiumjdbc_databasemetadata_data_definition_ignored_in_transactions"

    external deletes_are_detected : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_deletes_are_detected"

    external does_max_row_size_include_blobs : t -> bool =
      "cadmiumjdbc_databasemetadata_does_max_row_size_include_blobs"

    external get_attributes : t -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_attributes"

    external get_best_row_identifier : t -> string -> string -> string -> result_scope -> bool -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_best_row_identifier" "cadmiumjdbc_databasemetadata_get_best_row_identifier_native"

    external get_catalogs : t -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_catalogs"

    external get_catalog_separator : t -> string =
      "cadmiumjdbc_databasemetadata_get_catalog_separator"

    external get_catalog_term : t -> string =
      "cadmiumjdbc_databasemetadata_get_catalog_term"

    external get_client_info_properties : t -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_client_info_properties"

    external get_column_privileges : t -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_column_privileges"

    external get_columns : t -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_columns"

    external get_cross_reference : t -> string -> string -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_cross_reference" "cadmiumjdbc_databasemetadata_get_cross_reference_native"

    external get_database_major_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_database_major_version"

    external get_database_minor_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_database_minor_version"

    external get_database_product_name : t -> string =
      "cadmiumjdbc_databasemetadata_get_database_product_name"

    external get_database_product_version : t -> string =
      "cadmiumjdbc_databasemetadata_get_database_product_version"

    external get_default_transaction_isolation : t -> transaction_isolation =
      "cadmiumjdbc_databasemetadata_get_default_transaction_isolation"

    external get_driver_major_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_driver_major_version"

    external get_driver_minor_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_driver_minor_version"

    external get_driver_name : t -> string =
      "cadmiumjdbc_databasemetadata_get_driver_name"

    external get_driver_version : t -> string =
      "cadmiumjdbc_databasemetadata_get_driver_version"

    external get_exported_keys : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_exported_keys"

    external get_extra_name_characters : t -> string =
      "cadmiumjdbc_databasemetadata_get_extra_name_characters"

    external get_function_columns : t -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_function_columns"

    external get_functions : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_functions"

    external get_identifier_quote_string : t -> string =
      "cadmiumjdbc_databasemetadata_get_identifier_quote_string"

    external get_imported_keys : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_imported_keys"

    external get_index_info : t -> string -> string -> string -> bool -> bool -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_index_info" "cadmiumjdbc_databasemetadata_get_index_info_native"

    external get_jdbc_major_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_jdbc_major_version"

    external get_jdbc_minor_version : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_jdbc_minor_version"

    external get_max_binary_literal_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_binary_literal_length"

    external get_max_catalog_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_catalog_name_length"

    external get_max_char_literal_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_char_literal_length"

    external get_max_column_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_column_name_length"

    external get_max_columns_in_group_by : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_columns_in_group_by"

    external get_max_columns_in_index : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_columns_in_index"

    external get_max_columns_in_order_by : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_columns_in_order_by"

    external get_max_columns_in_select : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_columns_in_select"

    external get_max_columns_in_table : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_columns_in_table"

    external get_max_connections : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_connections"

    external get_max_cursor_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_cursor_name_length"

    external get_max_index_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_index_length"

    external get_max_procedure_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_procedure_name_length"

    external get_max_row_size : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_row_size"

    external get_max_schema_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_schema_name_length"

    external get_max_statement_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_statement_length"

    external get_max_statements : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_statements"

    external get_max_table_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_table_name_length"

    external get_max_tables_in_select : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_tables_in_select"

    external get_max_user_name_length : t -> int32 =
      "cadmiumjdbc_databasemetadata_get_max_user_name_length"

    external get_numeric_functions : t -> string =
      "cadmiumjdbc_databasemetadata_get_numeric_functions"

    external get_primary_keys : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_primary_keys"

    external get_procedure_columns : t -> string -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_procedure_columns"

    external get_procedures : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_procedures"

    external get_procedure_term : t -> string =
      "cadmiumjdbc_databasemetadata_get_procedure_term"

    external get_result_set_holdability : t -> result_set_holdability =
      "cadmiumjdbc_databasemetadata_get_result_set_holdability"

    external get_row_id_lifetime : t -> rowid_lifetime =
      "cadmiumjdbc_databasemetadata_get_row_id_lifetime"

    external get_schemas : t -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_schemas"

    external get_schemas_params : t -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_schemas_params"

    external get_schema_term : t -> string =
      "cadmiumjdbc_databasemetadata_get_schema_term"

    external get_search_string_escape : t -> string =
      "cadmiumjdbc_databasemetadata_get_search_string_escape"

    external get_sql_keywords : t -> string =
      "cadmiumjdbc_databasemetadata_get_sql_keywords"

    external get_sql_state_type : t -> sql_state_type =
      "cadmiumjdbc_databasemetadata_get_sql_state_type"

    external get_string_functions : t -> string =
      "cadmiumjdbc_databasemetadata_get_string_functions"

    external get_super_tables : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_super_tables"

    external get_super_types : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_super_types"

    external get_system_functions : t -> string =
      "cadmiumjdbc_databasemetadata_get_system_functions"

    external get_table_privileges : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_table_privileges"

    external get_tables : t -> string -> string -> string -> string list -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_tables"

    external get_table_types : t -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_table_types"

    external get_time_date_functions : t -> string =
      "cadmiumjdbc_databasemetadata_get_time_date_functions"

    external get_type_info : t -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_type_info"

    external get_udts : t -> string -> string -> string -> jdbc_type list -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_udts"

    external get_url : t -> string =
      "cadmiumjdbc_databasemetadata_get_url"

    external get_user_name : t -> string =
      "cadmiumjdbc_databasemetadata_get_user_name"

    external get_version_columns : t -> string -> string -> string -> ResultSet.t =
      "cadmiumjdbc_databasemetadata_get_version_columns"

    external inserts_are_detected : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_inserts_are_detected"

    external is_catalog_at_start : t -> bool =
      "cadmiumjdbc_databasemetadata_is_catalog_at_start"

    external is_read_only : t -> bool =
      "cadmiumjdbc_databasemetadata_is_read_only"

    external locators_update_copy : t -> bool =
      "cadmiumjdbc_databasemetadata_locators_update_copy"

    external null_plus_non_null_is_null : t -> bool =
      "cadmiumjdbc_databasemetadata_null_plus_non_null_is_null"

    external nulls_are_sorted_at_end : t -> bool =
      "cadmiumjdbc_databasemetadata_nulls_are_sorted_at_end"

    external nulls_are_sorted_at_start : t -> bool =
      "cadmiumjdbc_databasemetadata_nulls_are_sorted_at_start"

    external nulls_are_sorted_high : t -> bool =
      "cadmiumjdbc_databasemetadata_nulls_are_sorted_high"

    external nulls_are_sorted_low : t -> bool =
      "cadmiumjdbc_databasemetadata_nulls_are_sorted_low"

    external others_deletes_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_others_deletes_are_visible"

    external others_inserts_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_others_inserts_are_visible"

    external others_updates_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_others_updates_are_visible"

    external own_deletes_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_own_deletes_are_visible"

    external own_inserts_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_own_inserts_are_visible"

    external own_updates_are_visible : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_own_updates_are_visible"

    external stores_lower_case_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_lower_case_identifiers"

    external stores_lower_case_quoted_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_lower_case_quoted_identifiers"

    external stores_mixed_case_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_mixed_case_identifiers"

    external stores_mixed_case_quoted_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_mixed_case_quoted_identifiers"

    external stores_upper_case_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_upper_case_identifiers"

    external stores_upper_case_quoted_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_stores_upper_case_quoted_identifiers"

    external supports_alter_table_with_add_column : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_alter_table_with_add_column"

    external supports_alter_table_with_drop_column : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_alter_table_with_drop_column"

    external supports_ansi92_entry_level_sql : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_ansi92_entry_level_sql"

    external supports_ansi92_full_sql : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_ansi92_full_sql"

    external supports_ansi92_intermediate_sql : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_ansi92_intermediate_sql"

    external supports_batch_updates : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_batch_updates"

    external supports_catalogs_in_data_manipulation : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_catalogs_in_data_manipulation"

    external supports_catalogs_in_index_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_catalogs_in_index_definitions"

    external supports_catalogs_in_privilege_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_catalogs_in_privilege_definitions"

    external supports_catalogs_in_procedure_calls : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_catalogs_in_procedure_calls"

    external supports_catalogs_in_table_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_catalogs_in_table_definitions"

    external supports_column_aliasing : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_column_aliasing"

    external supports_convert : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_convert"

    external supports_convert_types : t -> jdbc_type -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_supports_convert_types"

    external supports_core_sql_grammar : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_core_sql_grammar"

    external supports_correlated_subqueries : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_correlated_subqueries"

    external supports_data_definition_and_data_manipulation_transactions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_data_definition_and_data_manipulation_transactions"

    external supports_data_manipulation_transactions_only : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_data_manipulation_transactions_only"

    external supports_different_table_correlation_names : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_different_table_correlation_names"

    external supports_expressions_in_order_by : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_expressions_in_order_by"

    external supports_extended_sql_grammar : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_extended_sql_grammar"

    external supports_full_outer_joins : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_full_outer_joins"

    external supports_get_generated_keys : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_get_generated_keys"

    external supports_group_by : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_group_by"

    external supports_group_by_beyond_select : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_group_by_beyond_select"

    external supports_group_by_unrelated : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_group_by_unrelated"

    external supports_integrity_enhancement_facility : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_integrity_enhancement_facility"

    external supports_like_escape_clause : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_like_escape_clause"

    external supports_limited_outer_joins : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_limited_outer_joins"

    external supports_minimum_sql_grammar : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_minimum_sql_grammar"

    external supports_mixed_case_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_mixed_case_identifiers"

    external supports_mixed_case_quoted_identifiers : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_mixed_case_quoted_identifiers"

    external supports_multiple_open_results : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_multiple_open_results"

    external supports_multiple_result_sets : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_multiple_result_sets"

    external supports_multiple_transactions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_multiple_transactions"

    external supports_named_parameters : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_named_parameters"

    external supports_non_nullable_columns : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_non_nullable_columns"

    external supports_open_cursors_across_commit : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_open_cursors_across_commit"

    external supports_open_cursors_across_rollback : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_open_cursors_across_rollback"

    external supports_open_statements_across_commit : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_open_statements_across_commit"

    external supports_open_statements_across_rollback : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_open_statements_across_rollback"

    external supports_order_by_unrelated : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_order_by_unrelated"

    external supports_outer_joins : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_outer_joins"

    external supports_positioned_delete : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_positioned_delete"

    external supports_positioned_update : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_positioned_update"

    external supports_result_set_concurrency : t -> result_set_type -> result_set_concurrency -> bool =
      "cadmiumjdbc_databasemetadata_supports_result_set_concurrency"

    external supports_result_set_holdability : t -> result_set_holdability -> bool =
      "cadmiumjdbc_databasemetadata_supports_result_set_holdability"

    external supports_result_set_type : t -> result_set_type -> bool =
      "cadmiumjdbc_databasemetadata_supports_result_set_type"

    external supports_savepoints : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_savepoints"

    external supports_schemas_in_data_manipulation : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_schemas_in_data_manipulation"

    external supports_schemas_in_index_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_schemas_in_index_definitions"

    external supports_schemas_in_privilege_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_schemas_in_privilege_definitions"

    external supports_schemas_in_procedure_calls : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_schemas_in_procedure_calls"

    external supports_schemas_in_table_definitions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_schemas_in_table_definitions"

    external supports_select_for_update : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_select_for_update"

    external supports_statement_pooling : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_statement_pooling"

    external supports_stored_functions_using_call_syntax : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_stored_functions_using_call_syntax"

    external supports_stored_procedures : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_stored_procedures"

    external supports_subqueries_in_comparisons : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_subqueries_in_comparisons"

    external supports_subqueries_in_exists : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_subqueries_in_exists"

    external supports_subqueries_in_ins : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_subqueries_in_ins"

    external supports_subqueries_in_quantifieds : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_subqueries_in_quantifieds"

    external supports_table_correlation_names : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_table_correlation_names"

    external supports_transaction_isolation_level : t -> transaction_isolation -> bool =
      "cadmiumjdbc_databasemetadata_supports_transaction_isolation_level"

    external supports_transactions : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_transactions"

    external supports_union : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_union"

    external supports_union_all : t -> bool =
      "cadmiumjdbc_databasemetadata_supports_union_all"

    external updates_are_detected : t -> jdbc_type -> bool =
      "cadmiumjdbc_databasemetadata_updates_are_detected"

    external uses_local_file_per_table : t -> bool =
      "cadmiumjdbc_databasemetadata_uses_local_file_per_table"

    external uses_local_files : t -> bool =
      "cadmiumjdbc_databasemetadata_uses_local_files"
  end

module Connection =
  struct
    type t

    external clear_warnings : t -> unit =
      "cadmiumjdbc_connection_clear_warnings"

    external close : t -> unit =
      "cadmiumjdbc_connection_close"

    external commit : t -> unit =
      "cadmiumjdbc_connection_commit"

    external create_array_of : t -> string -> Cadmium.java_object list -> Array.t =
      "cadmiumjdbc_connection_create_array_of"

    external create_blob : t -> Blob.t =
      "cadmiumjdbc_connection_create_blob"

    external create_clob : t -> Clob.t =
      "cadmiumjdbc_connection_create_clob"

    external create_nclob : t -> NClob.t =
      "cadmiumjdbc_connection_create_nclob"

    external create_sqlxml : t -> SQLXML.t =
      "cadmiumjdbc_connection_create_sqlxml"

    external create_statement : t -> Statement.t =
      "cadmiumjdbc_connection_create_statement"

    external create_statement_type_conc : t -> result_set_type -> result_set_concurrency -> Statement.t =
      "cadmiumjdbc_connection_create_statement_type_conc"

    external create_statement_type_conc_hold : t -> result_set_type -> result_set_concurrency -> result_set_holdability -> Statement.t =
      "cadmiumjdbc_connection_create_statement_type_conc_hold"

    external create_struct : t -> string -> Cadmium.java_object list -> Struct.t =
      "cadmiumjdbc_connection_create_struct"

    external get_auto_commit : t -> bool =
      "cadmiumjdbc_connection_get_auto_commit"

    external get_catalog : t -> string =
      "cadmiumjdbc_connection_get_catalog"

    external get_client_infos : t -> (string * string) list =
      "cadmiumjdbc_connection_get_client_infos"

    external get_client_info : t -> string -> string =
      "cadmiumjdbc_connection_get_client_info"

    external get_holdability : t -> result_set_holdability =
      "cadmiumjdbc_connection_get_holdability"

    external get_meta_data : t -> DatabaseMetaData.t =
      "cadmiumjdbc_connection_get_meta_data"

    external get_transaction_isolation : t -> transaction_isolation =
      "cadmiumjdbc_connection_get_transaction_isolation"

    external get_warnings : t -> sql_warning list =
      "cadmiumjdbc_connection_get_warnings"

    external is_closed : t -> bool =
      "cadmiumjdbc_connection_is_closed"

    external is_read_only : t -> bool =
      "cadmiumjdbc_connection_is_read_only"

    external is_valid : t -> int32 -> bool =
      "cadmiumjdbc_connection_is_valid"

    external native_sql : t -> string -> string =
      "cadmiumjdbc_connection_native_sql"

    external prepare_call : t -> string -> CallableStatement.t =
      "cadmiumjdbc_connection_prepare_call"

    external prepare_call_type_conc : t -> string -> result_set_type -> result_set_concurrency -> CallableStatement.t =
      "cadmiumjdbc_connection_prepare_call_type_conc"

    external prepare_call_type_conc_hold : t -> string -> result_set_type -> result_set_concurrency -> result_set_holdability -> CallableStatement.t =
      "cadmiumjdbc_connection_prepare_call_type_conc_hold"

    external prepare_statement : t -> string -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement"

    external prepare_statement_autogen_keys : t -> string -> statement_autogen_keys -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement_autogen_keys"

    external prepare_statement_indexes : t -> string -> int list -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement_indexes"

    external prepare_statement_type_conc : t -> string -> result_set_type -> result_set_concurrency -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement_type_conc"

    external prepare_statement_type_conc_hold : t -> string -> result_set_type -> result_set_concurrency -> result_set_holdability -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement_type_conc_hold"

    external prepare_statement_names : t -> string -> string list -> PreparedStatement.t =
      "cadmiumjdbc_connection_prepare_statement_names"

    external release_savepoint : t -> Savepoint.t -> unit =
      "cadmiumjdbc_connection_release_savepoint"

    external rollback : t -> unit =
      "cadmiumjdbc_connection_rollback"

    external rollback_to : t -> Savepoint.t -> unit =
      "cadmiumjdbc_connection_rollback_to"

    external set_auto_commit : t -> bool -> unit =
      "cadmiumjdbc_connection_set_auto_commit"

    external set_catalog : t -> string -> unit =
      "cadmiumjdbc_connection_set_catalog"

    external set_client_infos : t -> (string * string) list -> unit =
      "cadmiumjdbc_connection_set_client_infos"

    external set_client_info : t -> string -> string -> unit =
      "cadmiumjdbc_connection_set_client_info"

    external set_holdability : t -> result_set_holdability -> unit =
      "cadmiumjdbc_connection_set_holdability"

    external set_read_only : t -> bool -> unit =
      "cadmiumjdbc_connection_set_read_only"

    external set_savepoint : t -> Savepoint.t =
      "cadmiumjdbc_connection_set_savepoint"

    external set_savepoint_with_name : t -> string -> Savepoint.t =
      "cadmiumjdbc_connection_set_savepoint_with_name"

    external set_transaction_isolation : t -> transaction_isolation -> unit =
      "cadmiumjdbc_connection_set_transaction_isolation"
  end

module Driver =
  struct
    type property_info = {
        choices : string list;
        description : string;
        name : string;
        required : bool;
        value : string;
      }

    type t

    external accepts_url : t -> string -> bool =
      "cadmiumjdbc_driver_accepts_url"

    external connect : t -> string -> (string * string) list -> Connection.t =
      "cadmiumjdbc_driver_connect"

    external get_major_version : t -> int32 =
      "cadmiumjdbc_driver_get_major_version"

    external get_minor_version : t -> int32 =
      "cadmiumjdbc_driver_get_minor_version"

    external get_property_info : t -> string -> (string * string) list -> property_info list =
      "cadmiumjdbc_driver_get_property_info"

    external is_jdbc_compliant : t -> bool =
      "cadmiumjdbc_driver_is_jdbc_compliant"
  end

module DriverManager =
  struct
    external deregister_driver : Driver.t -> unit =
      "cadmiumjdbc_drivermanager_deregister_driver"

    external get_connection : string -> Connection.t =
      "cadmiumjdbc_drivermanager_get_connection"

    external get_connection_with_properties : string -> (string * string) list -> Connection.t =
      "cadmiumjdbc_drivermanager_get_connection_with_properties"

    external get_connection_with_user_passwd : string -> string -> string -> Connection.t =
      "cadmiumjdbc_drivermanager_get_connection_with_user_passwd"

    external get_driver : string -> Driver.t =
      "cadmiumjdbc_drivermanager_get_driver"

    external get_drivers : unit -> Driver.t list =
      "cadmiumjdbc_drivermanager_get_drivers"

    external get_login_timeout : unit -> int32 =
      "cadmiumjdbc_drivermanager_get_login_timeout"

    external print_log : string -> unit =
      "cadmiumjdbc_drivermanager_print_log"

    external register_driver : Driver.t -> unit =
      "cadmiumjdbc_drivermanager_register_driver"

    external set_login_timeout : int32 -> unit =
      "cadmiumjdbc_drivermanager_set_login_timeout"
  end

module Iterators =
  struct
    exception Done

    let get_next o = try o#next with Not_found -> raise Done

    let iter f o =
      try
        while true do
          f (get_next o)
        done
      with Done -> ()

    let iter_close f o =
      iter f o;
      o#close

    let iteri f o =
      let i = ref 0 in
      try
        while true do
          f !i (get_next o);
          incr i
        done
      with Done -> ()

    let iteri_close f o =
      iteri f o;
      o#close

    let map f o =
      let res = ref [] in
      (try
        while true do
          res := (f (get_next o)) :: !res
        done;
      with Done -> ());
      List.rev !res

    let map_close f o =
      let res = map f o in
      o#close;
      res

    let mapi f o =
      let i = ref 0 in
      let res = ref [] in
      (try
        while true do
          res := (f !i (get_next o)) :: !res;
          incr i
        done
      with Done -> ());
      List.rev !res

    let mapi_close f o =
      let res = mapi f o in
      o#close;
      res

    let fold f z o =
      let res = ref z in
      (try
        while true do
          res := f !res (get_next o)
        done
      with Done -> ());
      !res

    let fold_close f z o =
      let res = fold f z o in
      o#close;
      res

    let foldi f z o =
      let i = ref 0 in
      let res = ref z in
      (try
        while true do
          res := f !i !res (get_next o);
          incr i
        done
      with Done -> ());
      !res

    let foldi_close f z o =
      let res = foldi f z o in
      o#close;
      res

    let to_list o =
      map (fun x -> x) o

    let to_list_close o =
      let res = to_list o in
      o#close;
      res
  end

let () = (* to ensure that the Cadmium library is initialized *)
  ignore (Cadmium.is_available ())
