(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to Java elements as OCaml classes and objects. *)


(** {6 Types} *)

type 'c java_type = 'c Cadmium.java_type =
  | Boolean                   (** boolean Java type *)
  | Byte                      (** byte Java type (8-bit signed integer) *)
  | Char                      (** character Java type (16-bit unsigned integer) *)
  | Double                    (** double Java type (double-precision floating-point number) *)
  | Float                     (** float Java type (single-precision floating-point number) *)
  | Int                       (** int Java type (32-bit signed integer) *)
  | Long                      (** long Java type (64-bit signed integer) *)
  | Short                     (** short Java type (16-bit signed integer) *)
  | String                    (** string of Java characters (immutable) *)
  | Class of 'c               (** any (other) Java type *)
  | Void                      (** void/no return type *)
  | Array of 'c java_type     (** array, parameter is elements type *)
(** The type of Java types (primitives, arrays and bare objects). *)

type 'o java_value = 'o Cadmium.java_value =
  | Boolean_val of bool       (** boolean Java value *)
  | Byte_val of int           (** byte Java value (8-bit signed integer) *)
  | Char_val of int           (** character Java value (16-bit unsigned integer) *)
  | Double_val of float       (** double Java value (double-precision floating-point number) *)
  | Float_val of float        (** float Java value (single-precision floating-point number) *)
  | Int_val of int32          (** int Java value (32-bit signed integer) *)
  | Long_val of int64         (** long Java value (64-bit signed integer) *)
  | Short_val of int          (** short Java value (16-bit signed integer) *)
  | String_val of string      (** string of Java characters (immutable) *)
  | Object_val of 'o          (** any (other) Java value *)
(** The type of Java values (primitives and bare objects). *)

type java_modifier = Cadmium.java_modifier =
  | Abstract                  (** type / method modifier *)
  | Final                     (** type / method / field modifier *)
  | Interface                 (** type modifier *)
  | Native                    (** method modifier *)
  | Private                   (** visibility modifier *)
  | Protected                 (** visibility modifier *)
  | Public                    (** visibility modifier *)
  | Static                    (** for class-wide definitions (as opposed to instance-wide definitions) *)
  | StrictFP                  (** for strict floating-point operations *)
  | Synchronized              (** synchronization modifier *)
  | Transcient                (** field (serialization) modifier *)
  | Volatile                  (** field modifier *)
(** The type of Java modifiers (for classes, fields, methods, {i etc.}). *)


(** {6 Base functions} *)

val is_available : unit -> bool
(** Tests whether Cadmium primitives are available. *)

val java_char_of_char : char -> int
(** Converts a [char] into a Java char.*)

val char_of_java_char : int -> char
(** Converts a Java char into a [char].*)


(** {6 java.lang and java.lang.reflect} *)

(** Each of the [jObject], [jClass], [jConstructor], [jField], [jMethod],
   and [jArray] classes are "translations" of the {i java.lang.Object},
   {i java.lang.Class}, {i java.lang.reflect.Constructor},
   {i java.lang.reflect.Field}, {i java.lang.reflect.Method},
   and {i java.lang.reflect.Array} Java classes. These classes are wrappers
   around types and methods defined by [Cadmium] module.

   Translation from Java is straighforward, except concerning the following
   elements:
- [cd'this] return the underlying [Cadmium.java_object] instance;
- [cd'equal] and [cd'notequal] are equivalent to {i ==} and {i !=} operators;
- [cd'instanceof] is equivalent to {i instanceof} operator;
- [cd'isnull] tests object against {i null};
- [cd'cast] is equivalent to cast operator;
- [cd'to_TYPE_array] converts a Java array into an OCaml one (it is the
   responsibility of the programmer to call the method corresponding to the
   actual type of elements in the array);
- [`Cd'init x] and [`Cd'initObj y] variants value allows construction of an
   instance from a [Cadmium.java_object] value or from a [CadmiumObj.jObject]
   value (other variant used for constructors have names that make their
   meanings obvious);
- some method names are modified as keeping the original ones would result in
  name clashes. *)

class jObject :
  [< `Cd'init of Cadmium.java_object | `Cd'initObj of < cd'this : Cadmium.java_object; .. > | `Void | `Null | `String of string
   > `Cd'init ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_object
    method clone : jObject
    method equals : jObject -> bool
    method getClass : jClass
    method hashCode : int32
    method notify : unit
    method notifyAll : unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end
and jClass :
  [< `Cd'init of Cadmium.java_class | `Cd'initObj of < cd'this : Cadmium.java_class; .. > | `For_name of string > `Cd'init ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_class
    method clone : jObject
    method equals : jObject -> bool
    method getCanonicalName : string
    method getClass : jClass
    method getConstructor : jClass java_type list -> jConstructor
    method getConstructors : jConstructor list
    method getDeclaredConstructor : jClass java_type list -> jConstructor
    method getDeclaredConstructors : jConstructor list
    method getDeclaredField : string -> jField
    method getDeclaredFields : jField list
    method getDeclaredMethod : string -> jClass java_type list -> jMethod
    method getDeclaredMethods : jMethod list
    method getField : string -> jField
    method getFields : jField list
    method getInterfaces : jClass list
    method getMethod : string -> jClass java_type list -> jMethod
    method getMethods : jMethod list
    method getModifiers : java_modifier list
    method getName : string
    method getPackage : string
    method getSimpleName : string
    method getSuperclass : jClass option
    method hashCode : int32
    method isArray : bool
    method isInterface : bool
    method isPrimitive : bool
    method newInstance : jObject
    method notify : unit
    method notifyAll : unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end
and jConstructor :
  [< `Cd'init of Cadmium.java_constructor | `Cd'initObj of < cd'this : Cadmium.java_constructor; .. > > `Cd'init ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_constructor
    method clone : jObject
    method equals : jObject -> bool
    method getClass : jClass
    method getDeclaringClass : jClass
    method getExceptionTypes : jClass list
    method getModifiers : java_modifier list
    method getParameterTypes : jClass java_type list
    method hashCode : int32
    method newInstance : jObject java_value list -> jObject
    method notify : unit
    method notifyAll : unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end
and jField :
  [< `Cd'init of Cadmium.java_field | `Cd'initObj of < cd'this : Cadmium.java_field; .. > > `Cd'init ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_field
    method clone : jObject
    method equals : jObject -> bool
    method get : jObject -> jObject
    method getBoolean : jObject -> bool
    method getByte : jObject -> int
    method getChar : jObject -> int
    method getClass : jClass
    method getDeclaringClass : jClass
    method getDouble : jObject -> float
    method getFloat : jObject -> float
    method getInt : jObject -> int32
    method getLong : jObject -> int64
    method getModifiers : java_modifier list
    method getName : string
    method getShort : jObject -> int
    method hashCode : int32
    method notify : unit
    method notifyAll : unit
    method set : jObject -> jObject -> unit
    method setBoolean : jObject -> bool -> unit
    method setByte : jObject -> int -> unit
    method setChar : jObject -> int -> unit
    method setDouble : jObject -> float -> unit
    method setFloat : jObject -> float -> unit
    method setInt : jObject -> int32 -> unit
    method setLong : jObject -> int64 -> unit
    method setShort : jObject -> int -> unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end
and jMethod :
  [< `Cd'init of Cadmium.java_method | `Cd'initObj of < cd'this : Cadmium.java_method; .. > > `Cd'init ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_method
    method clone : jObject
    method equals : jObject -> bool
    method getClass : jClass
    method getDeclaringClass : jClass
    method getExceptionTypes : jClass list
    method getModifiers : java_modifier list
    method getName : string
    method getParameterTypes : jClass java_type list
    method getReturnType : jClass java_type
    method hashCode : int32
    method invoke : jObject -> jObject java_value list -> jObject java_value
    method notify : unit
    method notifyAll : unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end
and jArray :
  [< `Cd'init of Cadmium.java_object
   | `Cd'initObj of < cd'this : Cadmium.java_object; .. >
   | `New of jClass java_type * int
   | `New_dims of jClass java_type * int list
   | `From_bool_array of bool array
   | `From_byte_array of int array
   | `From_char_array of char array
   | `From_double_array of float array
   | `From_float_array of float array
   | `From_int_array of int32 array
   | `From_long_array of int64 array
   | `From_short_array of int array
   | `From_string_array of string array ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_object
    method cd'to_bool_array : bool array
    method cd'to_byte_array : int array
    method cd'to_char_array : char array
    method cd'to_double_array : float array
    method cd'to_float_array : float array
    method cd'to_int_array : int32 array
    method cd'to_long_array : int64 array
    method cd'to_short_array : int array
    method cd'to_string_array : string array
    method clone : jObject
    method equals : jObject -> bool
    method get : int -> jObject
    method getBoolean : int -> bool
    method getByte : int -> int
    method getChar : int -> int
    method getClass : jClass
    method getDouble : int -> float
    method getFloat : int -> float
    method getInt : int -> int32
    method getLength : int
    method getLong : int -> int64
    method getShort : int -> int
    method hashCode : int32
    method notify : unit
    method notifyAll : unit
    method set : int -> jObject -> unit
    method setBoolean : int -> bool -> unit
    method setByte : int -> int -> unit
    method setChar : int -> int -> unit
    method setDouble : int -> float -> unit
    method setFloat : int -> float -> unit
    method setInt : int -> int32 -> unit
    method setLong : int -> int64 -> unit
    method setShort : int -> int -> unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end


(** {6 Exceptions} *)

exception Java_exception of (string * string * Cadmium.java_object)
(** Encapsulates Java exceptions, arguments are:
   exception class, exception message and exception instance *)

val get_exception_instance : exn -> jObject
(** [get_exception_instance e] returns the actual exception instance if e is
   a [Java_exception] and raises [Invalid_argument] otherwise. *)

val print_stack_trace : out_channel -> exn -> unit
(** Prints the stack trace of a Java exception onto an output channel,
   printing the OCaml exception if it is not a Java one. *)

val throw_exception : jObject -> unit
(** [throw_exception e] throws [e] if it is an exception, raises
   [Invalid_argument] otherwise *)


(** {6 Access to Graphics elements} *)

val get_frame : unit -> jObject
(** Returns the frame used by the 'Graphics' module (can be {i null} if no frame
   is used by the module). Raises [Invalid_argument] if the 'Graphics' module
   has not been initialized. *)

val get_canvas : unit -> jObject
(** Returns the canvas used by the 'Graphics' module.
   Raises [Invalid_argument] if the 'Graphics' module has not been initialized. *)

val get_buffer : unit -> jObject
(** Returns the buffer used by the 'Graphics' module.
   Raises [Invalid_argument] if the 'Graphics' module has not been initialized. *)


(** {6 Callbacks / Proxies} *)

val create_interface : (string * (jClass java_type list) * jClass java_type) list -> jClass
(** [create_interface meths] returns the Java class corresponding to the newly
   compiled interface that contains the methods from [l]. [l] contains method
   definitions given by a name, a list of parameter types and a return type.
   The fully qualified name of the interface is auto-generated. *)

val create_named_interface : (string * (jClass java_type list) * jClass java_type) list -> string -> jClass
(** [create_named_interface meths name] returns the Java class corresponding
   to the newly compiled interface whose fully qualified name is [name] that
   contains the methods from [l]. [l] contains method definitions given by a
   name, a list of parameter types and a return type. *)

val register : jClass list -> < .. > -> jObject
(** {i cf}. [Cadmium.register]

   {b Warning:} parameters and as well as return type of
   {i java.lang.Object} type should use [jObject] type. *)

val register_transl : jClass list -> < .. > -> (jMethod * string) list -> jObject
(** {i cf}. [Cadmium.register_object_transl]

   {b Warning:} parameters and as well as return type of
   {i java.lang.Object} type should use [jObject] type. *)


(** {6 Synchronization} *)

val synchronized : jObject -> (unit -> unit) -> unit
(** [synchronized o f] executes [f ()] in a synchronized block over [o]. *)


(** {6 Applet} *)

type applet_parameter = string * string * string
(** Represents an applet parameter defined by its name, type, and description. *)

type applet_handler = {
    info : string; (** Data to be returned by {i java.applet.Applet.getAppletInfo()}. *)
    parameters : applet_parameter array; (** Data to be returned by {i java.applet.Applet.getParameterInfo()}. *)
    init : jObject -> unit; (** Callback for {i java.applet.Applet.init()}, parameter being applet instance. *)
    start : jObject -> unit; (** Callback for {i java.applet.Applet.start()}, parameter being applet instance. *)
    run : jObject -> unit; (** Called by the applet thread launched right after the end of {i start}. *)
    stop : jObject -> unit; (** Callback for {i java.applet.Applet.stop()}, parameter being applet instance. *)
    destroy : jObject -> unit; (** Callback for {i java.applet.Applet.destroy()}, parameter being applet instance. *)
  }
(** Applet handler defining applet behaviour. *)

val register_applet : applet_handler -> unit
(** Registers the passed handler defining applet behaviour,
   replacing any previously registered handler. *)

val get_applet_parameter : jObject -> string -> string
(** [get_applet_parameter applet name] returns the value of the parameter named
   [name] for the applet [applet]. Raises [Not_found] if the parameter is not
   set. *)


(** {6 Runtime} *)

(** The following functions allow access to elements of the Cadmium runtime.
   A {i code runner} is a Cadmium thread running OCaml code either natively
   or through interpretation. A {i context} represents the state of an OCaml
   program and is shared between the various threads of this program. Finally,
   an {i interpreter} is an object that runs OCaml bytecode. By using such an
   interpreter it is possible for a native Cadmium program to run OCaml
   bytecode.

   For more information about {i code runner}, {i context}, and {i interpreter}
   as well as related objects, one is advised to refer to the Cadmium
   javadoc. *)

type interpreter_parameters = {
    file : string; (** argv.(0) value, NOT bytecode file *)
    arguments : string array; (** other arguments to program *)
    init_stack_size : int; (** initial stack size (in elements) *)
    max_stack_size : int; (** maximum stack size (in elements) *)
    backtrace : bool; (** whether backtrace is activated *)
    exit_stops_jvm : bool; (** whether exit stops the whole JVM *)
    providers : string array; (** additional primitive providers as fully-qualified Java class names *)
    compile_dispatchers : bool; (** whether dispatchers should be compiled *)
    use_awt : bool; (** whether the [Graphics] frame should be AWT *)
    bare_canvas : bool; (** whether the [Graphics] component should be a bare canvas *)
    use_javax_sound : bool; (** whether javax.sound should be used for beeps *)
    use_jdbm : bool; (** whether jdbm should be used for dbm *)
    os : string; (** OS identifier (suggested values: {i Unix}, {i Cygwin}, {i Win32}, {i MacOS}, {i Cadmium}) *)
    unix_emulation : bool (** whether unix primitives should be emulated *)
  }
(** Interpreter parameters. *)

val is_native : unit -> bool
(** Tests whether the current code runner is native. *)

val get_coderunner : unit -> jObject
(** Returns the current code runner ({i fr.x9c.cadmium.kernel.CodeRunner}
   instance). *)

val get_context : unit -> jObject
(** Returns the current context ({i fr.x9c.cadmium.kernel.Context}
   instance). *)

val get_parameters : unit -> interpreter_parameters
(** Returns the current parameters. *)

class interpreter :
  [< `Params of interpreter_parameters * string * string * in_channel * out_channel * out_channel | `Cd'init of Cadmium.java_object | `Cd'initObj of < cd'this : Cadmium.java_object; .. > ] ->
  object
    method cd'cast : jClass -> jObject
    method cd'equal : jObject -> bool
    method cd'instanceof : jClass -> bool
    method cd'isnull : bool
    method cd'notequal : jObject -> bool
    method cd'this : Cadmium.java_object
    method clone : jObject
    method equals : jObject -> bool
    method execute : 'a
    method executeCallback : string -> 'a list -> 'b
    method getClass : jClass
    method getContext : jObject
    method getParameters : interpreter_parameters
    method hashCode : int32
    method notify : unit
    method notifyAll : unit
    method toString : string
    method wait : unit
    method wait'1 : int64 -> unit
    method wait'2 : int64 -> int32 -> unit
  end


(** {6 Scripts} *)

val get_binding : string -> 'a
(** Returns the value associated with the passed identifier.
   If the value is a Java primitive, primitive wrapper, or a string, it is
   automatically mapped to the corresponding OCaml value. Otherwise, it is
   presented as a jObject instance.
   Raises [Not_found] if such a value does not exist. *)
