(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** Access to SwiXml engine (http://www.swixml.org). *)

(** This module is a straightforward translation of {i org.swixml.SwingEngine}.
   Each of the functions may raise a [Cadmium.Java_exception] if the 
   underlying Java method throws an exception. *)


(** {6 Configuration} *)

val is_debug_mode : unit -> bool
(** Tests whether SwiXml is in debug or release mode. *)

val is_macosx : unit -> bool
(** Tests whether the underlying platform is MacOS X. *)

val is_macosx_supported : unit -> bool
(** Tests whether MacOS X support is enabled. *)

val set_macosx_support : bool -> unit
(** Enables or disables MacOS X support. *)

val set_default_locale : Cadmium.java_object -> unit
(** Sets the default locale. The parameter should be a {i java.util.Locale}
   instance. *)

val set_resource_bundle_name : string -> unit
(** Sets the default resource bundle name. *)

val set_app_frame : Cadmium.java_object -> unit
(** Sets the global application frame, used as the parent for child dialogs.
   The parameter should be a {i java.awt.Frame} instance. *)

val get_app_frame : unit -> Cadmium.java_object
(** Returns the global application frame. *)


(** {6 Engine construction} *)

type engine
(** Type for {i org.swixml.SwingEngine} instances. *)

val make_engine : unit -> engine
(** Construct an engine. *)

val make_engine_from_resource : string -> engine
(** Construct an engine from a Java resource (should contain XML data
   describing UI). *)

val make_engine_from_client : Cadmium.java_object -> engine
(** Construct an engine from a client information will be read/written from/to
   instance fields using reflection. One should refer to the SwiXml
   documentation for information on this process. *)


(** {6 Engine parameters} *)

val get_client : engine -> Cadmium.java_object
(** Returns the client object of an engine. *)

val set_classloader : engine -> Cadmium.java_object -> unit
(** Sets the classloader of an engine. The parameter should be a
   {i java.lang.Classloader} instance. *)

val get_classloader : engine -> Cadmium.java_object
(** Returns the classloader of an engine. *)

val set_locale : engine -> Cadmium.java_object -> unit
(** Sets the locale of an engine. The parameter should be a
   {i java.util.Locale} instance. *)

val set_resource_bundle : engine -> string -> unit
(** Sets the resource bundle of an engine. *)


(** {6 Rendering} *)

val render_from_file : engine -> string -> Cadmium.java_object
(** [render_from_file e f] renders the XML file [f] into a UI using engine [e]
   and returns the root component of the UI. *)

val render_from_url : engine -> string -> Cadmium.java_object
(** [render_from_url e u] renders the XML file located at [u] into a UI using
   engine [e] and returns the root component of the UI. *)

val render_from_resource : engine -> string -> Cadmium.java_object
(** [render_from_resource e r] renders the XML file designated by resource [r]
   into a UI using engine [e] and returns the root component of the UI. *)

val render_from_channel : engine -> in_channel -> Cadmium.java_object
(** [render_from_channel e c] renders the XML file read from [c] into a UI
   using engine [e] and returns the root component of the UI. *)


(** {6 Insertion} *)

val insert_from_url : engine -> string -> Cadmium.java_object -> unit
(** [insert_from_url e u o] inserts into [o] UI elements rendered from the XML
   file located at [u] using engine [e]. [o] should be a {i java.awt.Container}
   instance. *)

val insert_from_resource : engine -> string -> Cadmium.java_object -> unit
(** [insert_from_resource e r o] inserts into [o] UI elements rendered from the
   XML file designated by [r] using engine [e]. [o] should be a
   {i java.awt.Container} instance. *)

val insert_from_channel : engine -> in_channel -> Cadmium.java_object -> unit
(** [insert_from_channel e c o] inserts into [o] UI elements rendered from the
   XML file read from [c] using engine [e]. [o] should be a
   {i java.awt.Container} instance. *)


(** {6 Access to components} *)

val get_root_component : engine -> Cadmium.java_object
(** Returns the root component of the rendered UI of an engine. The returned
   object is an instance of {i java.awt.Container}. *)

val get_id_map : engine -> (string * Cadmium.java_object) list
(** Returns the identifier/component map of an engine. *)

val clear_id_map : engine -> unit
(** Clears the identifier/component map of an engine. *)

val get_all_components : engine -> Cadmium.java_object list
(** Returns the list of all components of an engine. *)

val get_descendants : engine -> Cadmium.java_object -> Cadmium.java_object list
(** [get_descendants e o] returns the list of all descendants of [o] in engine
   [e]. [o] should be a {i java.awt.Component} instance.*)

val find : engine -> string -> Cadmium.java_object
(** [find e id] looks up for a component whose identifier is [id], searching in
   engine [e]. Raises [Not_found] if such a component does not exist. *)

val forget : engine -> string -> unit
(** [forget e id] forget the component whose identifier is [id] in engine [e].
 *)

val cleanup : engine -> int
(** Forgets all undisplayable components of an engine. Returns the number of
   such forgotten components. *)

val set_action_listener : engine -> Cadmium.java_object -> Cadmium.java_object -> bool
(** [set_action_listener e cmp lst] sets action listener [lst] with component
   [cmp] in engine [e]. [lst] should be a {i java.awt.event.ActionListener}
   instance and [cmp] should be a {i java.awt.Component} instance. Returns
   [true] if registration was successful, [false] otherwise. *)

val set_action_listener_by_name : engine -> string -> Cadmium.java_object -> bool
(** Utility method, [set_action_listener_by_name e id lst] is equivalent to
   [set_action_listener e cmp lst] where [cmp] is the result of [find e id]. *)


(** {6 Miscellaneous} *)

val java_object_of_engine : engine -> Cadmium.java_object
(** "Converts" an engine into a bare Java object. *)
