(*
 * This file is part of Cadmium.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cadmium is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Cadmium is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

module SchemaFactory =
  struct
    type t

    external make : string -> t =
      "cadmiumxml_schemafactory_make"

    external set_feature : t -> string -> bool -> unit =
      "cadmiumxml_schemafactory_set_feature"

    external get_feature : t -> string -> bool =
      "cadmiumxml_schemafactory_get_feature"

    external set_property : t -> string -> Cadmium.java_object -> unit =
      "cadmiumxml_schemafactory_set_property"

    let set_property_obj x s o = set_property x s o#cd'this

    external get_property : t -> string -> Cadmium.java_object =
      "cadmiumxml_schemafactory_get_property"

    let get_property_obj x s = new CadmiumObj.jObject (`Cd'init (get_property x s))
  end


module Schema =
  struct
    type t

    external make_schema_from_file : SchemaFactory.t -> string -> t =
      "cadmiumxml_schema_from_file"

    external make_schema_from_url : SchemaFactory.t -> string -> t =
      "cadmiumxml_schema_from_url"
  end


module SAXFactory =
  struct
    type t

    external make : unit -> t =
      "cadmiumxml_saxfactory_make"

    external set_feature : t -> string -> bool -> unit =
      "cadmiumxml_saxfactory_set_feature"

    external get_feature : t -> string -> bool =
      "cadmiumxml_saxfactory_get_feature"

    external set_schema : t -> Schema.t -> unit =
      "cadmiumxml_saxfactory_set_schema"

    external get_schema : t -> Schema.t =
      "cadmiumxml_saxfactory_get_schema"

    external set_namespace_aware : t -> bool -> unit =
      "cadmiumxml_saxfactory_set_namespace_aware"

    external is_namespace_aware : t -> bool =
      "cadmiumxml_saxfactory_is_namespace_aware"

    external set_validating : t -> bool -> unit =
      "cadmiumxml_saxfactory_set_validating"

    external is_validating : t -> bool =
      "cadmiumxml_saxfactory_is_validating"

    external set_xinclude_aware : t -> bool -> unit =
      "cadmiumxml_saxfactory_set_xinclude_aware"

    external is_xinclude_aware : t -> bool =
      "cadmiumxml_saxfactory_is_xinclude_aware"
  end


module SAXParser =
  struct
    type t

    type parse_error = int * int * string

    type handler = {
        characters : string -> unit;
        ignorable_whitespace : string -> unit;
        start_document : unit -> unit;
        end_document : unit -> unit;
        start_element : string -> string -> string -> ((string * string) list) -> unit;
        end_element : string -> string -> string -> unit;
        start_prefix_mapping : string -> string -> unit;
        end_prefix_mapping : string -> unit;
        error : parse_error -> unit;
        fatal_error : parse_error -> unit;
        warning : parse_error -> unit;
      }

    let default_handler = {
        characters = (fun _ -> ());
        ignorable_whitespace = (fun _ -> ());
        start_document = (fun _ -> ());
        end_document = (fun _ -> ());
        start_element = (fun _ _ _ _ -> ());
        end_element = (fun _ _ _ -> ());
        start_prefix_mapping = (fun _ _ -> ());
        end_prefix_mapping = (fun _ -> ());
        error = (fun _ -> ());
        fatal_error = (fun _ -> ());
        warning = (fun _ -> ());
      }

    external make : SAXFactory.t -> t =
      "cadmiumxml_sax_make"

    external set_property : t -> string -> Cadmium.java_object -> unit =
      "cadmiumxml_sax_set_property"

    let set_property_obj x s o = set_property x s o#cd'this

    external get_property : t -> string -> Cadmium.java_object =
      "cadmiumxml_sax_get_property"

    let get_property_obj x s = new CadmiumObj.jObject (`Cd'init (get_property x s))

    external get_schema : t -> Schema.t =
      "cadmiumxml_sax_get_schema"

    external is_namespace_aware : t -> bool =
      "cadmiumxml_sax_is_namespace_aware"

    external is_validating : t -> bool =
      "cadmiumxml_sax_is_validating"

    external is_xinclude_aware : t -> bool =
      "cadmiumxml_sax_is_xinclude_aware"

    external parse_file : t -> handler -> string -> unit =
      "cadmiumxml_sax_parse_file"

    external parse_channel : t -> handler -> in_channel -> unit =
      "cadmiumxml_sax_parse_channel"

    external reset : t -> unit =
      "cadmiumxml_sax_reset"
  end


module Document =
  struct
    (* Types *)

    type attr

    type cdata

    type comment

    type document

    type document_type

    type element

    type entity

    type entity_reference

    type notation

    type processing_instruction

    type text

    type node =
      | Attr of attr
      | Cdata of cdata
      | Comment of comment
      | Document of document
      | Document_type of document_type
      | Element of element
      | Entity of entity
      | Entity_reference of entity_reference
      | Notation of notation
      | Processing_instruction of processing_instruction
      | Text of text


          (* Document functions *)

    external get_document_element : document -> element =
      "cadmiumxml_document_get_document_element"

    external get_xml_encoding : document -> string =
      "cadmiumxml_document_get_xml_encoding"

    external set_strict_error_checking : document -> bool -> unit =
      "cadmiumxml_document_set_strict_error_checking"

    external is_strict_error_checking : document -> bool =
      "cadmiumxml_document_is_strict_error_checking"

    external set_xml_standalone : document -> bool -> unit =
      "cadmiumxml_document_set_xml_standalone"

    external is_xml_standalone : document -> bool =
      "cadmiumxml_document_is_xml_standalone"

    external set_xml_version : document -> bool -> unit =
      "cadmiumxml_document_set_xml_version"

    external get_xml_version : document -> string =
      "cadmiumxml_document_get_xml_version"

    external normalize_document : document -> unit =
      "cadmiumxml_document_normalize_document"

    let node_of_document d = Document d


        (* Attribute functions *)

    external get_attr_parent : attr -> element =
      "cadmiumxml_document_get_attr_parent"

    external get_attr_name : attr -> string =
      "cadmiumxml_document_get_attr_name"

    external get_attr_value : attr -> string =
      "cadmiumxml_document_get_attr_value"

    external set_attr_value : attr -> string -> unit =
      "cadmiumxml_document_set_attr_value"

    let node_of_attr a = Attr a


        (* CData functions *)

    external get_cdata_parent : cdata -> node =
      "cadmiumxml_document_get_cdata_parent"

    external get_cdata : cdata -> string =
      "cadmiumxml_document_get_cdata"

    external set_cdata : cdata -> string -> unit =
      "cadmiumxml_document_set_cdata"

    let node_of_cdata c = Cdata c


        (* Comment functions *)

    external get_comment_parent : comment -> node =
      "cadmiumxml_document_get_comment_parent"

    external get_comment : comment -> string =
      "cadmiumxml_document_get_comment"

    external set_comment : comment -> string -> unit =
      "cadmiumxml_document_set_comment"

    let node_of_comment c = Comment c


        (* Document type functions *)

    external get_document_type_parent : document_type -> node =
      "cadmiumxml_document_get_document_type_parent"

    external get_document_type_name : document_type -> string =
      "cadmiumxml_document_get_document_type_name"

    external get_document_type_public_id : document_type -> string =
      "cadmiumxml_document_get_document_type_public_id"

    external get_document_type_system_id : document_type -> string =
      "cadmiumxml_document_get_document_type_system_id"

    let node_of_document_type d = Document_type d


        (* Element functions *)

    external get_element_parent : element -> node =
      "cadmiumxml_document_get_element_parent"

    external get_attribute : element -> string -> string =
      "cadmiumxml_document_get_attribute"

    external get_attribute_ns : element -> string -> string -> string =
      "cadmiumxml_document_get_attribute_ns"

    external get_attribute_node : element -> string -> attr =
      "cadmiumxml_document_get_attribute_node"

    external get_attribute_node_ns : element -> string -> string -> attr =
      "cadmiumxml_document_get_attribute_node_ns"

    external get_elements_by_tag_name : element -> string -> element list =
      "cadmiumxml_document_get_elements_by_tag_name"

    external get_elements_by_tag_name_ns : element -> string -> string -> element list =
      "cadmiumxml_document_get_elements_by_tag_name_ns"

    external get_tag_name : element -> string =
      "cadmiumxml_document_get_tag_name"

    external has_attribute : element -> string -> bool =
      "cadmiumxml_document_has_attribute"

    external has_attribute_ns : element -> string -> string -> bool =
      "cadmiumxml_document_has_attribute_ns"

    external remove_attribute : element -> string -> unit =
      "cadmiumxml_document_remove_attribute"

    external remove_attribute_ns : element -> string -> string -> unit =
      "cadmiumxml_document_remove_attribute_ns"

    external remove_attribute_node : element -> attr -> unit =
      "cadmiumxml_document_remove_attribute_node"

    external set_attribute : element -> string -> string -> unit =
      "cadmiumxml_document_set_attribute"

    external set_attribute_ns : element -> string -> string -> string -> unit =
      "cadmiumxml_document_set_attribute_ns"

    let node_of_element e = Element e


        (* Entity functions *)

    external get_entity_parent : entity -> node =
      "cadmiumxml_document_get_entity_parent"

    external get_entity_notation_name : entity -> string =
      "cadmiumxml_document_get_entity_notation_name"

    external get_entity_public_id : entity -> string =
      "cadmiumxml_document_get_entity_public_id"

    external get_entity_system_id : entity -> string =
      "cadmiumxml_document_get_entity_system_id"

    let node_of_entity e = Entity e


        (* Notation functions *)

    external get_notation_parent : notation -> node =
      "cadmiumxml_document_get_notation_parent"

    external get_notation_public_id : notation -> string =
      "cadmiumxml_document_get_notation_public_id"

    external get_notation_system_id : notation -> string =
      "cadmiumxml_document_get_notation_system_id"

    let node_of_notation n = Notation n


        (* Processing instruction functions *)

    external get_processing_instruction_parent : processing_instruction -> node =
      "cadmiumxml_document_get_processing_instruction_parent"

    external get_processing_instruction_target : processing_instruction -> string =
      "cadmiumxml_document_get_processing_instruction_target"

    external get_processing_instruction_data : processing_instruction -> string =
      "cadmiumxml_document_get_processing_instruction_data"

    external set_processing_instruction_data : processing_instruction -> string -> unit =
      "cadmiumxml_document_set_processing_instruction_data"

    let node_of_processing_instruction pi = Processing_instruction pi


        (* Text functions *)

    external get_text_parent : text -> node =
      "cadmiumxml_document_get_text_parent"

    external get_text : text -> string =
      "cadmiumxml_document_get_text"

    external set_text : text -> string -> unit =
      "cadmiumxml_document_set_text"

    let node_of_text t = Text t


        (* Node functions *)

    external get_parent : node -> node =
      "cadmiumxml_document_get_parent"

    external get_document : node -> document =
      "cadmiumxml_document_get_document"

    external append_attr_child : node -> string -> attr =
      "cadmiumxml_document_append_attr_child"

    external append_attr_ns_child : node -> string -> string -> attr =
      "cadmiumxml_document_append_attr_ns_child"

    external append_cdata_child : node -> string -> cdata =
      "cadmiumxml_document_append_cdata_child"

    external append_comment_child : node -> string -> comment =
      "cadmiumxml_document_append_comment_child"

    external append_element_child : node -> string -> element =
      "cadmiumxml_document_append_element_child"

    external append_element_ns_child : node -> string -> string -> element =
      "cadmiumxml_document_append_element_ns_child"

    external append_entity_reference_child : node -> string -> entity_reference =
      "cadmiumxml_document_append_entity_reference_child"

    external insert_attr_child_before : node -> node -> string -> attr =
      "cadmiumxml_document_insert_attr_child_before"

    external insert_attr_ns_child_before : node -> node -> string -> string -> attr =
      "cadmiumxml_document_insert_attr_ns_child_before"

    external insert_cdata_child_before : node -> node -> string -> cdata =
      "cadmiumxml_document_insert_cdata_child_before"

    external insert_comment_child_before : node -> node -> string -> comment =
      "cadmiumxml_document_insert_comment_child_before"

    external insert_element_child_before : node -> node -> string -> element =
      "cadmiumxml_document_insert_element_child_before"

    external insert_element_ns_child_before : node -> node -> string -> string -> element =
      "cadmiumxml_document_insert_element_ns_child_before"

    external insert_entity_reference_child_before : node -> node -> string -> entity_reference =
      "cadmiumxml_document_insert_entity_reference_child_before"

    external get_child_nodes : node -> node list =
      "cadmiumxml_document_get_child_nodes"

    external get_first_child : node -> node =
      "cadmiumxml_document_get_first_child"

    external get_last_child : node -> node =
      "cadmiumxml_document_get_last_child"

    external get_previous_sibling : node -> node =
      "cadmiumxml_document_get_previous_sibling"

    external get_next_sibling : node -> node =
      "cadmiumxml_document_get_next_sibling"

    external remove_child : node -> node -> unit =
      "cadmiumxml_document_remove_child"

    external replace_child : node -> node -> node -> unit =
      "cadmiumxml_document_replace_child"
  end


module DocumentFactory =
  struct
    type t

    external make : unit -> t =
      "cadmiumxml_documentfactory_make"

    external set_attribute : t -> string -> Cadmium.java_object -> unit =
      "cadmiumxml_documentfactory_set_attribute"

    let set_attribute_obj x s o = set_attribute x s o#cd'this

    external get_attribute : t -> string -> Cadmium.java_object =
      "cadmiumxml_documentfactory_get_attribute"

    let get_attribute_obj x s = new CadmiumObj.jObject (`Cd'init (get_attribute x s))

    external set_feature : t -> string -> bool -> unit =
      "cadmiumxml_documentfactory_set_feature"

    external get_feature : t -> string -> bool =
      "cadmiumxml_documentfactory_get_feature"

    external set_schema : t -> Schema.t -> unit =
      "cadmiumxml_documentfactory_set_schema"

    external get_schema : t -> Schema.t =
      "cadmiumxml_documentfactory_get_schema"

    external set_coalescing : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_coalescing"

    external is_coalescing : t -> bool =
      "cadmiumxml_documentfactory_is_coalescing"

    external set_expand_entity_references : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_expand_entity_references"

    external is_expand_entity_references : t -> bool =
      "cadmiumxml_documentfactory_is_expand_entity_references"

    external set_ignoring_comments : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_ignoring_comments"

    external is_ignoring_comments : t -> bool =
      "cadmiumxml_documentfactory_is_ignoring_comments"

    external set_ignoring_element_content_whitespace : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_ignoring_element_content_whitespace"

    external is_ignoring_element_content_whitespace : t -> bool =
      "cadmiumxml_documentfactory_is_ignoring_element_content_whitespace"

    external set_namespace_aware : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_namespace_aware"

    external is_namespace_aware : t -> bool =
      "cadmiumxml_documentfactory_is_namespace_aware"

    external set_validating : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_validating"

    external is_validating : t -> bool =
      "cadmiumxml_documentfactory_is_validating"

    external set_xinclude_aware : t -> bool -> unit =
      "cadmiumxml_documentfactory_set_xinclude_aware"

    external is_xinclude_aware : t -> bool =
      "cadmiumxml_documentfactory_is_xinclude_aware"
  end


module DocumentBuilder =
  struct
    type t

    external make : DocumentFactory.t -> t =
      "cadmiumxml_documentbuilder_make"

    external get_schema : t -> Schema.t =
      "cadmiumxml_documentbuilder_get_schema"

    external is_namespace_aware : t -> bool =
      "cadmiumxml_documentbuilder_is_namespace_aware"

    external is_validating : t -> bool =
      "cadmiumxml_documentbuilder_is_validating"

    external is_xinclude_aware : t -> bool =
      "cadmiumxml_documentbuilder_is_xinclude_aware"

    external new_document : t -> Document.document =
      "cadmiumxml_documentbuilder_new_document"

    external parse_file : t -> string -> Document.document =
      "cadmiumxml_documentbuilder_parse_file"

    external parse_channel : t -> in_channel -> Document.document =
      "cadmiumxml_documentbuilder_parse_channel"

    external reset : t -> unit =
      "cadmiumxml_documentbuilder_reset"
  end


module Tree =
  struct
    type t = Element of element | Cdata of string | Text of string
    and element = { tag : string; attributes : (string * string) list; children : t list; }

    external of_document : Document.document -> t =
      "cadmiumxml_tree_of_document"
  end


module TransformerFactory =
  struct
    type t

    external make : unit -> t =
      "cadmiumxml_transformerfactory_make"

    external set_attribute : t -> string -> Cadmium.java_object -> unit =
      "cadmiumxml_transformerfactory_set_attribute"

    let set_attribute_obj x s o = set_attribute x s o#cd'this

    external get_attribute : t -> string -> Cadmium.java_object =
      "cadmiumxml_transformerfactory_get_attribute"

    let get_attribute_obj x s = new CadmiumObj.jObject (`Cd'init (get_attribute x s))

    external set_feature : t -> string -> bool -> unit =
      "cadmiumxml_transformerfactory_set_feature"

    external get_feature : t -> string -> bool =
      "cadmiumxml_transformerfactory_get_feature"
  end


module Transformer =
  struct
    type t

    external make_id : TransformerFactory.t -> t =
      "cadmiumxml_transformer_make_id"

    external make_from_file : TransformerFactory.t -> string -> t =
      "cadmiumxml_transformer_make_from_file"

    external make_from_channel : TransformerFactory.t -> in_channel -> t =
      "cadmiumxml_transformer_make_from_channel"

    external clear_parameters : t -> unit =
      "cadmiumxml_transformer_clear_parameters"

    external set_parameter : t -> string -> Cadmium.java_object -> unit =
      "cadmiumxml_transformer_set_parameter"

    let set_parameter_obj x s o = set_parameter x s o#cd'this

    external get_parameter : t -> string -> Cadmium.java_object =
      "cadmiumxml_transformer_get_parameter"

    let get_parameter_obj x s = new CadmiumObj.jObject (`Cd'init (get_parameter x s))

    external set_output_property : t -> string -> string -> unit =
      "cadmiumxml_transformer_set_output_property"

    external get_output_property : t -> string -> string =
      "cadmiumxml_transformer_get_output_property"

    external transform_file_to_file : t -> string -> string -> unit =
      "cadmiumxml_transformer_transform_file_to_file"

    external transform_file_to_channel : t -> string -> out_channel -> unit =
      "cadmiumxml_transformer_transform_file_to_channel"

    external transform_channel_to_file : t -> in_channel -> string -> unit =
      "cadmiumxml_transformer_transform_channel_to_file"

    external transform_channel_to_channel : t -> in_channel -> out_channel -> unit =
      "cadmiumxml_transformer_transform_channel_to_channel"

    external reset : t -> unit =
      "cadmiumxml_transformer_reset"
  end


module XPathFactory =
  struct
    type t

    external make : unit -> t =
      "cadmiumxml_xpathfactory_make"

    external set_feature : t -> string -> bool -> unit =
      "cadmiumxml_xpathfactory_set_feature"

    external get_feature : t -> string -> bool =
      "cadmiumxml_xpathfactory_get_feature"
  end


module XPath =
  struct
    type t

    type expr

    external make : XPathFactory.t -> t =
      "cadmiumxml_xpath_make"

    external compile : t -> string -> expr =
      "cadmiumxml_xpath_compile"

    external evaluate_file : t -> string -> string -> string =
      "cadmiumxml_xpath_evaluate_file"

    external evaluate_channel : t -> string -> in_channel -> string =
      "cadmiumxml_xpath_evaluate_channel"

    external evaluate_document : t -> string -> Document.document -> string =
      "cadmiumxml_xpath_evaluate_document"

    external expr_evaluate_file : expr -> string -> string =
      "cadmiumxml_xpath_expr_evaluate_file"

    external expr_evaluate_channel : expr -> in_channel -> string =
      "cadmiumxml_xpath_expr_evaluate_channel"

    external expr_evaluate_document : expr -> Document.document -> string =
      "cadmiumxml_xpath_expr_evaluate_document"

    external reset : t -> unit =
      "cadmiumxml_xpath_reset"
  end


let () = (* to ensure that the Cadmium library is initialized *)
  ignore (Cadmium.is_available ())
